import plistlib
from contextlib import suppress

import pytest

from rpcclient.exceptions import NoSuchPreferenceError

pytestmark = pytest.mark.darwin
DOMAIN = 'rpcserver'
USERNAME = 'kCFPreferencesAnyUser'
PLIST_PATH = f'/Library/Preferences/{DOMAIN}.plist'


@pytest.fixture()
def tmp_preference(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    client.preferences.cf.set_dict({'key1': 'value1'}, DOMAIN, USERNAME)
    client.preferences.cf.sync(DOMAIN, USERNAME)
    client.fs.write_file(PLIST_PATH, plistlib.dumps({'key1': 'value1'}))
    try:
        yield
    finally:
        try:
            client.preferences.cf.clear(DOMAIN, USERNAME)
        except NoSuchPreferenceError:
            pass
        client.preferences.cf.sync(DOMAIN, USERNAME)


def test_cf_get_keys(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    assert 'key1' in client.preferences.cf.get_keys(DOMAIN, USERNAME)


def test_cf_get_keys_invalid_preference(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with pytest.raises(NoSuchPreferenceError):
        client.preferences.cf.get_keys('com.apple.invalid_preference_for_sure', USERNAME)


def test_cf_get_value(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    assert 'value1' == client.preferences.cf.get_value('key1', DOMAIN, USERNAME)


def test_cf_get_dict(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    assert 'value1' == client.preferences.cf.get_dict(DOMAIN, USERNAME)['key1']


def test_cf_set(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    client.preferences.cf.set('key2', {'hey': 'you'}, DOMAIN, USERNAME)
    assert client.preferences.cf.get_value('key2', DOMAIN, USERNAME) == {'hey': 'you'}


def test_cf_remove(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    client.preferences.cf.remove('key1', DOMAIN, USERNAME)
    assert client.preferences.cf.get_value('key1', DOMAIN, USERNAME) is None


def test_cf_set_dict(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    client.preferences.cf.get_value('key1', DOMAIN, USERNAME)
    client.preferences.cf.set_dict({'b': 5}, DOMAIN, USERNAME)
    assert client.preferences.cf.get_dict(DOMAIN, USERNAME) == {'b': 5}


def test_cf_update_dict(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    update_contents = {'a': 5}
    expected_dict = client.preferences.cf.get_dict(DOMAIN, USERNAME)
    expected_dict.update(update_contents)
    client.preferences.cf.update_dict(update_contents, DOMAIN, USERNAME)
    assert client.preferences.cf.get_dict(DOMAIN, USERNAME) == expected_dict


def test_cf_clear(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    client.preferences.cf.clear(DOMAIN, USERNAME)
    with pytest.raises(NoSuchPreferenceError):
        assert not client.preferences.cf.get_dict(DOMAIN, USERNAME)


def test_sc_get_keys(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    keys = client.preferences.sc.get_keys(PLIST_PATH)
    assert 'key1' in keys


def test_sc_get_dict(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    dict_ = client.preferences.sc.get_dict(PLIST_PATH)
    assert 'key1' in dict_


def test_sc_object_keys(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.preferences.sc.open(PLIST_PATH) as pref:
        keys = pref.keys
    assert 'key1' in keys


def test_sc_object_set(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.preferences.sc.open(PLIST_PATH) as pref:
        pref.set('key2', 'value2')
    assert client.preferences.sc.get_dict(PLIST_PATH)['key2'] == 'value2'


def test_sc_object_set_dict(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.preferences.sc.open(PLIST_PATH) as pref:
        pref.set_dict({'hey': 'you'})
    assert client.preferences.sc.get_dict(PLIST_PATH) == {'hey': 'you'}


def test_sc_object_update_dict(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.preferences.sc.open(PLIST_PATH) as pref:
        pref.update_dict({'hey': 'you'})
    dict_ = client.preferences.sc.get_dict(PLIST_PATH)
    assert dict_['key1'] == 'value1'
    assert dict_['hey'] == 'you'


def test_sc_object_remove(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.preferences.sc.open(PLIST_PATH) as pref:
        pref.remove('key1')
    assert 'key1' not in client.preferences.sc.get_dict(PLIST_PATH)


def test_sc_object_get(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.preferences.sc.open(PLIST_PATH) as pref:
        val = pref.get('key1')
    assert val == 'value1'


def test_sc_object_get_dict(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.preferences.sc.open(PLIST_PATH) as pref:
        dict_ = pref.get_dict()
    assert dict_ == {'key1': 'value1'}


def test_sc_object_clear(client, tmp_preference):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.preferences.sc.open(PLIST_PATH) as pref:
        pref.clear()
    assert not client.preferences.sc.get_dict(PLIST_PATH)


class TestCustomDomain:
    @pytest.fixture(autouse=True)
    def clear_domain(self, client):
        with suppress(NoSuchPreferenceError):
            # if from some reason this domain already exist, empty it
            client.preferences.cf.clear(DOMAIN)
            assert client.preferences.cf.get_keys(DOMAIN) is None

    def test_set_dict(self, client):
        test_dict = {'KEY1': 'VALUE1', 'KEY2': 'VALUE2'}
        client.preferences.cf.set_dict(test_dict, DOMAIN)
        assert test_dict == client.preferences.cf.get_dict(DOMAIN)

    def test_remove(self, client):
        test_dict = {'KEY1': 'VALUE1', 'KEY2': 'VALUE2'}
        client.preferences.cf.set_dict(test_dict, DOMAIN)
        client.preferences.cf.remove('KEY2', DOMAIN)
        test_dict.pop('KEY2')
        assert test_dict == client.preferences.cf.get_dict(DOMAIN)


def test_sc_preferences(client):
    with client.preferences.sc.open(DOMAIN) as pref:
        if pref.get_dict() != {}:
            # if from some reason this domain already exist, empty it
            pref.clear()
            assert pref.get_dict() == {}

        # test set a full dictionary
        test_dict = {'KEY1': 'VALUE1', 'KEY2': 'VALUE2'}
        pref.set_dict(test_dict)
        assert test_dict == pref.get_dict()

        # test set remove a single value
        pref.remove('KEY2')
        test_dict.pop('KEY2')
        assert test_dict == pref.get_dict()

        # remove out test data and verify it works
        pref.clear()
        assert pref.get_dict() == {}



import logging
from typing import List, Mapping

from rpcclient.darwin.consts import IOPMUserActiveType
from rpcclient.exceptions import BadReturnValueError

logger = logging.getLogger(__name__)


class PowerAssertion:
    def __init__(self, client, id: int):
        self._client = client
        self.id = id

    def __enter__(self) -> 'PowerAssertion':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._client.symbols.IOPMAssertionRelease(self.id)


class Power:
    """ Power utils """

    def __init__(self, client):
        self._client = client

    def declare_user_activity(self, name: str, type_: IOPMUserActiveType) -> PowerAssertion:
        """ Declares that the user is active on the system """
        with self._client.safe_malloc(8) as p_assertion_id:
            err = self._client.symbols.IOPMAssertionDeclareUserActivity(self._client.cf(name), type_, p_assertion_id)
            if err != 0:
                raise BadReturnValueError('IOPMAssertionCreateWithProperties() failed')
            return PowerAssertion(self._client, p_assertion_id[0])

    def declare_network_client_activity(self, name: str) -> PowerAssertion:
        """
        A convenience function for handling remote network clients; this is a wrapper for holding
        kIOPMAssertNetworkClientActive
        """
        with self._client.safe_malloc(8) as p_assertion_id:
            err = self._client.symbols.IOPMDeclareNetworkClientActivity(self._client.cf(name), p_assertion_id)
            if err != 0:
                raise BadReturnValueError('IOPMAssertionCreateWithProperties() failed')
            return PowerAssertion(self._client, p_assertion_id[0])

    def create_assertion(self, name: str, type_: str, reason: str = None) -> PowerAssertion:
        properties = {'AssertName': name, 'AssertType': type_}
        if reason is not None:
            properties['HumanReadableReason'] = reason
        with self._client.safe_malloc(8) as p_assertion_id:
            err = self._client.symbols.IOPMAssertionCreateWithProperties(self._client.cf(properties), p_assertion_id)
            if err != 0:
                raise BadReturnValueError('IOPMAssertionCreateWithProperties() failed')
            return PowerAssertion(self._client, p_assertion_id[0])

    def copy_assertions_by_process(self) -> Mapping[int, Mapping]:
        """ Returns a dictionary listing all assertions, grouped by their owning process """
        with self._client.safe_malloc(8) as p_assertions:
            if self._client.symbols.IOPMCopyAssertionsByProcess(p_assertions) != 0:
                raise BadReturnValueError('IOPMCopyAssertionsByProcess() failed')
            assertions = p_assertions[0]
        result = {}
        key_enumerator = assertions.objc_call('keyEnumerator')
        while True:
            pid_object = key_enumerator.objc_call('nextObject')
            if pid_object == 0:
                break
            pid = pid_object.py()
            result[pid] = assertions.objc_call('objectForKey:', pid_object).py()
        return result

    def copy_scheduled_power_events(self) -> List[Mapping]:
        """ List all scheduled system power events """
        return self._client.symbols.IOPMCopyScheduledPowerEvents().py()

    def copy_assertions_status(self) -> Mapping[str, int]:
        """ Returns a list of available assertions and their system-wide levels """
        with self._client.safe_malloc(8) as result:
            if 0 != self._client.symbols.IOPMCopyAssertionsStatus(result):
                raise BadReturnValueError('IOPMCopyAssertionsStatus() failed')
            return result[0].py()

    def reboot(self) -> None:
        if self._client.symbols.reboot().c_int64 == -1:
            raise BadReturnValueError()

    def sleep(self) -> None:
        """
        Enter sustem sleep

        See: https://gist.github.com/atr000/416796
        """
        with self._client.safe_malloc(8) as p_master:
            err = self._client.symbols.IOMasterPort(self._client.symbols.bootstrap_port[0], p_master)
            if err != 0:
                raise BadReturnValueError(f'IOMasterPort didnt work, err is {err}')
            pmcon = self._client.symbols.IOPMFindPowerManagement(p_master[0])
            if pmcon == 0:
                raise BadReturnValueError('IOPMFindPowerManagement coudlnt establish connection')
            err = self._client.symbols.IOPMSleepSystem(pmcon)
            if err != 0:
                raise BadReturnValueError(f'IOPMSleepSystem didnt work. err is {err}')



from pathlib import Path

from rpcclient.protocol import DEFAULT_PORT

LAUNCHD_PID = 1
LAUNCHD_PATH = '/sbin/launchd'


def test_list_sanity(client):
    processes = client.processes.list()
    assert len(processes) > 2  # at least launchd and us should be running
    for p in processes:
        if p.pid == LAUNCHD_PID:
            assert p.path == LAUNCHD_PATH


def test_get_process_by_listening_port(client):
    # there should only be one process listening on this port and that's us
    worker_process = client.processes.get_by_pid(client.pid)
    assert client.processes.get_processes_by_listening_port(DEFAULT_PORT)[0].pid == worker_process.ppid


def test_process_object(client):
    server = client.processes.get_self()
    assert server.pid > 0
    assert len(server.regions) > 0
    assert len(server.images) > 0
    assert Path(server.images[0].path).resolve() == Path(server.path).resolve()
    fds = server.fds
    assert fds[0].fd == 0
    assert fds[1].fd == 1
    assert fds[2].fd == 2



import logging

import click
import coloredlogs

from rpcclient.client_factory import create_local, create_tcp_client
from rpcclient.protocol import DEFAULT_PORT

coloredlogs.install(level=logging.DEBUG)

logging.getLogger('asyncio').disabled = True
logging.getLogger('parso').disabled = True
logging.getLogger('parso.cache').disabled = True
logging.getLogger('parso.python.diff').disabled = True
logging.getLogger('blib2to3.pgen2.driver').disabled = True
logging.getLogger('humanfriendly.prompts').disabled = True


@click.command()
def rpclocal() -> None:
    """ connect to local machine """
    create_local().interactive()


@click.command()
@click.argument('hostname')
@click.option('-p', '--port', type=click.INT, default=DEFAULT_PORT)
@click.option('-r', '--rebind-symbols', is_flag=True, help='reload all symbols upon connection')
@click.option('-l', '--load-all-libraries', is_flag=True, help='load all libraries')
def rpcclient(hostname: str, port: int, rebind_symbols: bool, load_all_libraries: bool):
    """ connect to remote host """
    client = create_tcp_client(hostname, port=port)
    if load_all_libraries:
        client.load_all_libraries(rebind_symbols=False)
    if rebind_symbols:
        client.rebind_symbols()
    client.interactive()


if __name__ == '__main__':
    rpcclient()



import ast
import builtins
import json
import logging
import plistlib
import typing
from collections import namedtuple
from dataclasses import dataclass
from functools import lru_cache
from typing import Mapping

from cached_property import cached_property
from construct import Int8ul, Int32ul, Int64sl, Int64ul
from tqdm import tqdm, trange

from rpcclient.client import Client
from rpcclient.darwin import objective_c_class
from rpcclient.darwin.bluetooth import Bluetooth
from rpcclient.darwin.common import CfSerializable
from rpcclient.darwin.consts import CFPropertyListFormat, CFPropertyListMutabilityOptions, kCFAllocatorDefault
from rpcclient.darwin.core_graphics import CoreGraphics
from rpcclient.darwin.darwin_lief import DarwinLief
from rpcclient.darwin.fs import DarwinFs
from rpcclient.darwin.hid import Hid
from rpcclient.darwin.ioregistry import IORegistry
from rpcclient.darwin.keychain import Keychain
from rpcclient.darwin.location import Location
from rpcclient.darwin.media import DarwinMedia
from rpcclient.darwin.network import DarwinNetwork
from rpcclient.darwin.objective_c_symbol import ObjectiveCSymbol
from rpcclient.darwin.power import Power
from rpcclient.darwin.preferences import Preferences
from rpcclient.darwin.processes import DarwinProcesses
from rpcclient.darwin.structs import utsname
from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.darwin.syslog import Syslog
from rpcclient.darwin.time import Time
from rpcclient.darwin.xpc import Xpc
from rpcclient.exceptions import CfSerializationError, GettingObjectiveCClassError, MissingLibraryError
from rpcclient.protocol import arch_t, cmd_type_t, protocol_message_t
from rpcclient.structs.consts import RTLD_NOW
from rpcclient.symbol import Symbol
from rpcclient.symbols_jar import SymbolsJar

IsaMagic = namedtuple('IsaMagic', 'mask value')
ISA_MAGICS = [
    # ARM64
    IsaMagic(mask=0x000003f000000001, value=0x000001a000000001),
    # X86_64
    IsaMagic(mask=0x001f800000000001, value=0x001d800000000001),
]
# Mask for tagged pointer, from objc-internal.h
OBJC_TAG_MASK = (1 << 63)

FRAMEWORKS_PATH = '/System/Library/Frameworks'
PRIVATE_FRAMEWORKS_PATH = '/System/Library/PrivateFrameworks'
LIB_PATH = '/usr/lib'

FRAMEWORKS_BLACKLIST = (
    'PowerlogLiteOperators.framework', 'PowerlogCore.framework', 'PowerlogHelperdOperators.framework',
    'PowerlogFullOperators.framework', 'PowerlogAccounting.framework', 'JavaVM.framework', 'ActionKit.framework',
    'DashBoard.framework', 'CoverSheet.framework', 'StoreKitMacHelper.framework', 'ReplayKitMacHelper.framework',
    'PassKitMacHelper.framework')

logger = logging.getLogger(__name__)


@dataclass
class DyldImage:
    name: str
    base_address: int


class DarwinClient(Client):
    def __init__(self, sock, sysname: str, arch: arch_t, create_socket_cb: typing.Callable):
        super().__init__(sock, sysname, arch, create_socket_cb)
        self._dlsym_global_handle = -2  # RTLD_GLOBAL
        self._init_process_specific()

    def _init_process_specific(self):
        super(DarwinClient, self)._init_process_specific()

        if 0 == self.dlopen("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation", RTLD_NOW):
            raise MissingLibraryError('failed to load CoreFoundation')

        self.fs = DarwinFs(self)
        self.preferences = Preferences(self)
        self.processes = DarwinProcesses(self)
        self.media = DarwinMedia(self)
        self.ioregistry = IORegistry(self)
        self.location = Location(self)
        self.xpc = Xpc(self)
        self.syslog = Syslog(self)
        self.time = Time(self)
        self.hid = Hid(self)
        self.lief = DarwinLief(self)
        self.bluetooth = Bluetooth(self)
        self.core_graphics = CoreGraphics(self)
        self.keychain = Keychain(self)
        self.network = DarwinNetwork(self)
        self.power = Power(self)
        self.loaded_objc_classes = []
        self._NSPropertyListSerialization = self.objc_get_class('NSPropertyListSerialization')
        self._CFNullTypeID = self.symbols.CFNullGetTypeID()

    def interactive(self, additional_namespace: typing.Mapping = None):
        if additional_namespace is None:
            additional_namespace = {}
        additional_namespace['CFSTR'] = self.cf
        super().interactive(additional_namespace=additional_namespace)

    @property
    def images(self) -> typing.List[DyldImage]:
        m = []
        for i in range(self.symbols._dyld_image_count()):
            module_name = self.symbols._dyld_get_image_name(i).peek_str()
            base_address = self.symbols._dyld_get_image_header(i)
            m.append(
                DyldImage(module_name, base_address)
            )
        return m

    @cached_property
    def uname(self):
        with self.safe_calloc(utsname.sizeof()) as uname:
            assert 0 == self.symbols.uname(uname)
            return utsname.parse_stream(uname)

    @cached_property
    def is_idevice(self):
        return self.uname.machine.startswith('i')

    @property
    def roots(self) -> typing.List[str]:
        """ get a list of all accessible darwin roots when used for lookup of files/preferences/... """
        return ['/', '/var/root']

    def showobject(self, object_address: Symbol) -> Mapping:
        message = protocol_message_t.build({
            'cmd_type': cmd_type_t.CMD_SHOWOBJECT,
            'data': {'address': object_address},
        })
        with self._protocol_lock:
            self._sock.sendall(message)
            response_len = Int64sl.parse(self._recvall(Int64sl.sizeof()))
            response = self._recvall(response_len)
        return json.loads(response)

    def showclass(self, class_address: Symbol) -> Mapping:
        message = protocol_message_t.build({
            'cmd_type': cmd_type_t.CMD_SHOWCLASS,
            'data': {'address': class_address},
        })
        with self._protocol_lock:
            self._sock.sendall(message)
            response_len = Int64sl.parse(self._recvall(Int64sl.sizeof()))
            response = self._recvall(response_len)
        return json.loads(response)

    def get_class_list(self) -> typing.Mapping[str, objective_c_class.Class]:
        message = protocol_message_t.build({
            'cmd_type': cmd_type_t.CMD_GET_CLASS_LIST,
            'data': b'',
        })
        result = {}
        with self._protocol_lock:
            self._sock.sendall(message)
            count = Int32ul.parse(self._recvall(Int32ul.sizeof()))
            for _ in trange(count):
                name_len = Int8ul.parse(self._recvall(Int8ul.sizeof()))
                try:
                    name = self._recvall(name_len).decode()
                except UnicodeDecodeError:
                    self._recvall(Int64ul.sizeof())
                    continue

                class_ = objective_c_class.Class(self, self.symbol(Int64ul.parse(self._recvall(Int64ul.sizeof()))),
                                                 lazy=True)
                result[name] = class_
        return result

    def symbol(self, symbol: int):
        """ at a symbol object from a given address """
        return DarwinSymbol.create(symbol, self)

    def decode_cf(self, symbol: Symbol) -> CfSerializable:
        if self.symbols.CFGetTypeID(symbol) == self._CFNullTypeID:
            return None

        with self.safe_malloc(8) as p_error:
            p_error[0] = 0
            objc_data = self._NSPropertyListSerialization.dataWithPropertyList_format_options_error_(
                symbol, CFPropertyListFormat.kCFPropertyListBinaryFormat_v1_0, 0, p_error)
            if p_error[0] != 0:
                raise CfSerializationError()
        if objc_data == 0:
            return None
        count = self.symbols.CFDataGetLength(objc_data)
        result = plistlib.loads(self.symbols.CFDataGetBytePtr(objc_data).peek(count))
        objc_data.objc_call('release')
        return result

    def cf(self, o: CfSerializable) -> DarwinSymbol:
        """ construct a CFObject from a given python object """
        if o is None:
            return self.symbols.kCFNull[0]

        plist_bytes = plistlib.dumps(o, fmt=plistlib.FMT_BINARY)
        plist_objc_bytes = self.symbols.CFDataCreate(kCFAllocatorDefault, plist_bytes, len(plist_bytes))
        with self.safe_malloc(8) as p_error:
            p_error[0] = 0
            result = self._NSPropertyListSerialization.propertyListWithData_options_format_error_(
                plist_objc_bytes, CFPropertyListMutabilityOptions.kCFPropertyListMutableContainersAndLeaves, 0, p_error)
            if p_error[0] != 0:
                raise CfSerializationError()
            return result

    def objc_symbol(self, address) -> ObjectiveCSymbol:
        """
        Get objc symbol wrapper for given address
        :param address:
        :return: ObjectiveC symbol object
        """
        return ObjectiveCSymbol.create(int(address), self)

    @lru_cache(maxsize=None)
    def objc_get_class(self, name: str):
        """
        Get ObjC class object
        :param name:
        :return:
        """
        return objective_c_class.Class.from_class_name(self, name)

    def is_objc_type(self, symbol: DarwinSymbol) -> bool:
        """
        Test if a given symbol represents an objc object
        :param symbol:
        :return:
        """
        class_info = self.processes.get_self().get_symbol_class_info(symbol)
        if class_info == 0:
            return False
        return 'ObjC' == class_info.objc_call('typeName').py()

    def _add_global(self, name: str, value) -> None:
        super()._add_global(name, value)
        globals()[name] = value

    def _ipython_run_cell_hook(self, info):
        """
        Enable lazy loading for symbols
        :param info: IPython's CellInf4o object
        """
        super()._ipython_run_cell_hook(info)

        for node in ast.walk(ast.parse(info.raw_cell)):
            if not isinstance(node, ast.Name):
                # we are only interested in names
                continue

            if node.id in locals():
                continue

            global_var = globals().get(node.id)
            if global_var is not None:
                if isinstance(global_var, objective_c_class.Class) and global_var.name == '':
                    # reload lazy classes right before actual use
                    global_var.reload()
                continue

            if node.id in dir(builtins):
                continue

            if not hasattr(SymbolsJar, node.id):
                # ignore SymbolsJar properties
                try:
                    symbol = self.objc_get_class(node.id)
                except GettingObjectiveCClassError:
                    pass
                else:
                    self._add_global(
                        node.id,
                        symbol
                    )

    def rebind_symbols(self, populate_global_scope=True) -> None:
        logger.debug('rebinding symbols')
        self.loaded_objc_classes.clear()

        # enumerate all loaded objc classes
        for name, class_ in self.get_class_list().items():
            self.loaded_objc_classes.append(name)
            if populate_global_scope:
                self._add_global(
                    name,
                    class_
                )

    def load_framework(self, name: str) -> DarwinSymbol:
        lib = self.dlopen(f'{FRAMEWORKS_PATH}/{name}.framework/{name}', RTLD_NOW)
        if lib == 0:
            lib = self.dlopen(f'{PRIVATE_FRAMEWORKS_PATH}/{name}.framework/{name}', RTLD_NOW)
        if lib == 0:
            raise MissingLibraryError(f'failed to load {name}')

    def load_all_libraries(self, rebind_symbols=True) -> None:
        logger.debug(f'loading frameworks: {FRAMEWORKS_PATH}')
        self._load_frameworks(FRAMEWORKS_PATH)
        logger.debug(f'loading frameworks: {PRIVATE_FRAMEWORKS_PATH}')
        self._load_frameworks(PRIVATE_FRAMEWORKS_PATH)

        logger.debug(f'loading libraries: {LIB_PATH}')
        for filename in tqdm(self.fs.listdir(LIB_PATH)):
            if not filename.endswith('.dylib'):
                continue
            self.dlopen(f'{LIB_PATH}/{filename}', RTLD_NOW)

        if rebind_symbols:
            self.rebind_symbols()

    def _load_frameworks(self, frameworks_path: str):
        for filename in tqdm(self.fs.listdir(frameworks_path)):
            if filename in FRAMEWORKS_BLACKLIST:
                continue
            if 'SpringBoard' in filename or 'UI' in filename:
                continue
            self.dlopen(f'{frameworks_path}/{filename}/{filename.split(".", 1)[0]}', RTLD_NOW)



import gc


def test_allocate_file_fd_context_manager(client, tmp_path):
    # make sure when the test starts, all previous Allocated references are freed
    gc.collect()
    fds_count = len(client.processes.get_by_pid(client.pid).fds)
    with client.fs.open(tmp_path / 'test', 'w'):
        assert fds_count + 1 == len(client.processes.get_by_pid(client.pid).fds)
    assert fds_count == len(client.processes.get_by_pid(client.pid).fds)


def test_allocate_file_fd_explicit_deallocate(client, tmp_path):
    # make sure when the test starts, all previous Allocated references are freed
    gc.collect()
    fds_count = len(client.processes.get_by_pid(client.pid).fds)
    fd = client.fs.open(tmp_path / 'test', 'w')
    assert fds_count + 1 == len(client.processes.get_by_pid(client.pid).fds)
    fd.deallocate()
    assert fds_count == len(client.processes.get_by_pid(client.pid).fds)



import pytest

pytestmark = pytest.mark.darwin


def test_modules(client):
    """
    :param rpcclient.client.Client client:
    """
    assert '/usr/lib/libSystem.B.dylib' in [module.name for module in client.images]


def test_uname(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.uname.sysname == 'Darwin'
    assert 'machine' in client.uname


def test_get_class_list(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    assert len(client.get_class_list()) > 0


def test_load_framework(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    assert client.load_framework('Foundation') != 0


def test_load_all_libraries(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    original_count = len(client.images)
    client.load_all_libraries()
    assert len(client.images) > original_count



from collections import namedtuple

from rpcclient.exceptions import RpcFailedLaunchingAppError

ScreenLockStatus = namedtuple('ScreenLockStatus', ['lock', 'passcode'])


class SpringBoard:
    def __init__(self, client):
        self._client = client

    def get_spring_board_server_port(self) -> int:
        return self._client.symbols.SBSSpringBoardServerPort()

    def launch_application(self, bundle_identifier: str) -> None:
        """ launch application using SpringBoardServices """
        err = self._client.symbols.SBSLaunchApplicationWithIdentifier(self._client.cf(bundle_identifier), 0)
        if err != 0:
            raise RpcFailedLaunchingAppError(
                f'SBSLaunchApplicationWithIdentifier failed with: error code {err} - '
                f'{self._client.symbols.SBSApplicationLaunchingErrorString(err).py()}')

    def get_screen_lock_status(self) -> ScreenLockStatus:
        """ get lockscreen and passcode status using SpringBoardServices """
        server_port = self.get_spring_board_server_port()
        with self._client.safe_malloc(8) as p_is_lock:
            with self._client.safe_malloc(8) as p_is_passcode:
                p_is_lock[0] = 0
                p_is_passcode[0] = 0
                self._client.symbols.SBGetScreenLockStatus(server_port, p_is_lock, p_is_passcode)
                return ScreenLockStatus(p_is_lock[0] == 1, p_is_passcode[0] == 1)

    def open_sensitive_url_and_unlock(self, url: str, unlock: bool = True) -> None:
        """ open default application according to url scheme """
        screen_lock_status = self.get_screen_lock_status()
        if not unlock and screen_lock_status.lock:
            if screen_lock_status.passcode:
                raise RpcFailedLaunchingAppError(
                    'cannot open url while screen is locked with passcode. you must unlock device first')
            raise RpcFailedLaunchingAppError('cannot open url while screen is locked, use unlock=True parameter')
        cf_url_ref = self._client.objc_get_class('NSURL').URLWithString_(self._client.cf(url))
        if not self._client.symbols.SBSOpenSensitiveURLAndUnlock(cf_url_ref, unlock):
            raise RpcFailedLaunchingAppError('SBSOpenSensitiveURLAndUnlock failed')



import logging
import os
import subprocess
import tempfile
import typing
from pathlib import Path
from socket import socket
from zipfile import ZipFile

import requests

from rpcclient.client import Client
from rpcclient.exceptions import FailedToConnectError, InvalidServerVersionMagicError
from rpcclient.ios.client import IosClient
from rpcclient.linux.client import LinuxClient
from rpcclient.macos.client import MacosClient
from rpcclient.protocol import DEFAULT_PORT, SERVER_MAGIC_VERSION, protocol_handshake_t

PROJECT_URL = 'https://github.com/doronz88/rpc-project/archive/refs/heads/master.zip'
RPCSERVER_SUBDIR = 'rpc-project-master/src/rpcserver'
HOMEDIR = Path('~/.rpc-project').expanduser()

logger = logging.getLogger(__name__)


def recvall(sock, size: int) -> bytes:
    buf = b''
    while size:
        chunk = sock.recv(size)
        if not chunk:
            raise FailedToConnectError()
        size -= len(chunk)
        buf += chunk
    return buf


def create_local() -> typing.Union[Client, IosClient, MacosClient, LinuxClient]:
    if not HOMEDIR.exists():
        with tempfile.TemporaryDirectory() as temp_dir:
            local_zip = Path(temp_dir) / 'master.zip'
            local_zip.write_bytes(requests.get(PROJECT_URL).content)
            ZipFile(local_zip).extractall(HOMEDIR)
    os.chdir(HOMEDIR / RPCSERVER_SUBDIR)
    cwd = os.getcwd()

    assert 0 == os.system('make')

    server = socket()
    server.bind(('127.0.0.1', 0))
    available_port = server.getsockname()[1]
    server.close()

    proc = subprocess.Popen(['./rpcserver', '-p', str(available_port)], cwd=cwd)
    logger.info(f'server launched at pid {proc.pid}')

    while True:
        try:
            return create_tcp_client('127.0.0.1', available_port)
        except FailedToConnectError:
            pass


def create_tcp_client(hostname: str, port: int = DEFAULT_PORT) -> \
        typing.Union[Client, IosClient, MacosClient, LinuxClient]:
    def tcp_connect() -> socket:
        sock = socket()
        try:
            sock.connect((hostname, port))
        except ConnectionRefusedError as e:
            # wrap in our own exception
            raise FailedToConnectError() from e
        return sock

    return create_client(tcp_connect)


def create_client(create_socket_cb: typing.Callable) -> typing.Union[Client, IosClient, MacosClient, LinuxClient]:
    sock = create_socket_cb()
    handshake = protocol_handshake_t.parse(recvall(sock, protocol_handshake_t.sizeof()))

    if handshake.magic != SERVER_MAGIC_VERSION:
        raise InvalidServerVersionMagicError(f'got {handshake.magic:x} instead of {SERVER_MAGIC_VERSION:x}')

    sysname = handshake.sysname.lower()
    machine = handshake.machine.lower()
    arch = handshake.arch

    logging.info(f'connection uname.sysname: {sysname} uname.machine: {machine}')

    if sysname == 'darwin':
        if machine.startswith('iphone'):
            return IosClient(sock, sysname, arch, create_socket_cb)
        else:
            return MacosClient(sock, sysname, arch, create_socket_cb)
    elif sysname == 'linux':
        return LinuxClient(sock, sysname, arch, create_socket_cb)

    return Client(sock, sysname, arch, create_socket_cb)



from collections import namedtuple
from functools import partial
from pathlib import Path
from typing import Mapping, Optional

from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import ObjectiveCLexer

from rpcclient.darwin import objc
from rpcclient.exceptions import GettingObjectiveCClassError
from rpcclient.symbols_jar import SymbolsJar

Ivar = namedtuple('Ivar', 'name type_ offset')


class Class:
    """
    Wrapper for ObjectiveC Class object.
    """

    def __init__(self, client, class_object=0, class_data: Mapping = None, lazy=False):
        """
        :param rpcclient.darwin.client.DarwinClient client: Darwin client.
        :param rpcclient.darwin.objective_c_symbol.Symbol class_object:
        """
        self._client = client
        self._class_object = class_object
        self.protocols = []
        self.ivars = []
        self.properties = []
        self.methods = []
        self.name = ''
        self.super = None
        if not lazy:
            if class_data is None:
                self.reload()
            else:
                self._load_class_data(class_data)

    @staticmethod
    def from_class_name(client, class_name: str):
        """
        Create ObjectiveC Class from given class name.
        :param rpcclient.darwin.client.DarwinClient client: Darwin client.
        :param class_name: Class name.
        """
        class_object = client.symbols.objc_getClass(class_name)
        class_symbol = Class(client, class_object)
        if class_symbol.name != class_name:
            raise GettingObjectiveCClassError()
        return class_symbol

    @staticmethod
    def sanitize_name(name: str):
        """
        Sanitize python name to ObjectiveC name.
        """
        if name.startswith('_'):
            name = '_' + name[1:].replace('_', ':')
        else:
            name = name.replace('_', ':')
        return name

    def reload(self):
        """
        Reload class object data.
        Should be used whenever the class layout changes (for example, during method swizzling)
        """
        objc_class = self._class_object if self._class_object else self._client.symbols.objc_getClass(self.name)
        class_description = self._client.showclass(objc_class)

        self.super = Class(self._client, class_description['super']) if class_description['super'] else None
        self.name = class_description['name']
        self.protocols = class_description['protocols']
        self.ivars = [
            Ivar(name=ivar['name'], type_=ivar['type'], offset=ivar['offset'])
            for ivar in class_description['ivars']
        ]
        self.properties = [
            objc.Property(name=prop['name'], attributes=objc.convert_encoded_property_attributes(prop['attributes']))
            for prop in class_description['properties']
        ]
        self.methods = [
            objc.Method.from_data(method, self._client) for method in class_description['methods']
        ]

    def show(self, dump_to: Optional[str] = None):
        """
        Print to terminal the highlighted class description.
        :param dump_to: directory to dump.
        """
        formatted = str(self)
        print(highlight(formatted, ObjectiveCLexer(), TerminalTrueColorFormatter(style='native')))

        if dump_to is None:
            return
        (Path(dump_to) / f'{self.name}.m').expanduser().write_text(formatted)

    def objc_call(self, sel: str, *args, **kwargs):
        """
        Invoke a selector on the given class object.
        :param sel: Selector name.
        :return: whatever the selector returned as a symbol.
        """
        return self._class_object.objc_call(sel, *args, **kwargs)

    def get_method(self, name: str):
        """
        Get a specific method implementation.
        :param name: Method name.
        :return: Method.
        """
        for method in self.methods:
            if method.name == name:
                return method

    def iter_supers(self):
        """
        Iterate over the super classes of the class.
        """
        sup = self.super
        while sup is not None:
            yield sup
            sup = sup.super

    def _load_class_data(self, data: Mapping):
        self._class_object = self._client.symbol(data['address'])
        self.super = Class(self._client, data['super']) if data['super'] else None
        self.name = data['name']
        self.protocols = data['protocols']
        self.ivars = [Ivar(name=ivar['name'], type_=ivar['type'], offset=ivar['offset']) for ivar in data['ivars']]
        self.properties = data['properties']
        self.methods = data['methods']

    @property
    def symbols_jar(self) -> SymbolsJar:
        """ Get a SymbolsJar object for quick operations on all methods """
        jar = SymbolsJar.create(self._client)

        for m in self.methods:
            jar[f'[{self.name} {m.name}]'] = m.address

        return jar

    @property
    def bundle_path(self) -> Path:
        return Path(self._client.symbols.objc_getClass('NSBundle')
                    .objc_call('bundleForClass:', self._class_object).objc_call('bundlePath').py())

    def __dir__(self):
        result = set()

        for method in self.methods:
            if method.is_class:
                result.add(method.name.replace(':', '_'))

        for sup in self.iter_supers():
            for method in sup.methods:
                if method.is_class:
                    result.add(method.name.replace(':', '_'))

        result.update(list(super(Class, self).__dir__()))
        return list(result)

    def __str__(self):
        protocol_buf = f'<{",".join(self.protocols)}>' if self.protocols else ''

        if self.super is not None:
            buf = f'@interface {self.name}: {self.super.name} {protocol_buf}\n'
        else:
            buf = f'@interface {self.name} {protocol_buf}\n'

        # Add ivars
        buf += '{\n'
        for ivar in self.ivars:
            buf += f'\t{ivar.type_} {ivar.name}; // 0x{ivar.offset:x}\n'
        buf += '}\n'

        # Add properties
        for prop in self.properties:
            buf += f'@property ({",".join(prop.attributes.list)}) {prop.attributes.type_} {prop.name};\n'

            if prop.attributes.synthesize is not None:
                buf += f'@synthesize {prop.name} = {prop.attributes.synthesize};\n'

        # Add methods
        for method in self.methods:
            buf += str(method)

        buf += '@end'
        return buf

    def __repr__(self):
        return f'<objC Class "{self.name}">'

    def __getitem__(self, item):
        for method in self.methods:
            if method.name == item:
                if method.is_class:
                    return partial(self.objc_call, item)
                else:
                    raise AttributeError(f'{self.name} class has an instance method named {item}, '
                                         f'not a class method')

        for sup in self.iter_supers():
            for method in sup.methods:
                if method.name == item:
                    if method.is_class:
                        return partial(self.objc_call, item)
                    else:
                        raise AttributeError(f'{self.name} class has an instance method named {item}, '
                                             f'not a class method')

        raise AttributeError(f''''{self.name}' class has no attribute {item}''')

    def __getattr__(self, item: str):
        return self[self.sanitize_name(item)]



from construct import Adapter, Aligned, Array, BitsInteger, BitStruct, Bytes, Computed, CString, Default, Enum, \
    FlagsEnum, GreedyRange, Hex, Int8ul, Int16sl, Int16ub, Int16ul, Int32sl, Int32ub, Int32ul, Int64sl, Int64ul, \
    LazyArray, Octet, PaddedString, Padding, Seek, Struct, Switch, Tell, Union, this

from rpcclient.structs.consts import AF_INET, AF_INET6, AF_UNIX
from rpcclient.structs.generic import UNIX_PATH_MAX, gid_t, in_addr, long, mode_t, short, st_flags, time_t, u_char, \
    u_int32_t, u_short, uid_t, uint8_t, uint32_t, uint64_t

MAXPATHLEN = 1024
_SYS_NAMELEN = 256

utsname = Struct(
    'sysname' / PaddedString(_SYS_NAMELEN, 'utf8'),
    'nodename' / PaddedString(_SYS_NAMELEN, 'utf8'),
    'release' / PaddedString(_SYS_NAMELEN, 'utf8'),
    'version' / PaddedString(_SYS_NAMELEN, 'utf8'),
    'machine' / PaddedString(_SYS_NAMELEN, 'utf8'),
)

pid_t = Int32ul
exitcode_t = Int32sl
ino_t = Int32ul
dev_t = Int32ul
off_t = Int64ul
nlink_t = Int16ul
blkcnt_t = Int64ul
blksize_t = Int32ul
ino64_t = Int64ul
fsid_t = Array(2, Int32sl)
mach_port_t = Int64ul
io_name_t = PaddedString(1024, 'utf8')
io_object_t = Int64ul
vm_prot_t = Int32ul
vm_inherit_t = Int32ul
boolean_t = Int32ul
memory_object_offset_t = Int64ul
vm_behavior_t = Int32ul
mach_vm_address_t = Int64ul
mach_vm_size_t = Int64ul
integer_t = Int32sl

timespec = Struct(
    'tv_sec' / long,
    'tv_nsec' / long,
)
dirent32 = Struct(
    'd_ino' / ino_t,
    'd_offset' / Int16ul,
    'd_reclen' / Int8ul,
    'd_namelen' / Int8ul,
    'd_name' / PaddedString(this.d_namelen, 'utf8'),
)
dirent64 = Struct(
    'd_ino' / ino64_t,
    'd_offset' / Int64ul,
    'd_reclen' / Int16ul,
    'd_namelen' / Int16ul,
    'd_type' / Int8ul,
    'd_name' / PaddedString(this.d_namelen, 'utf8'),
)
stat32 = Struct(
    'st_dev' / dev_t,  # device inode resides on
    'st_ino' / ino_t,  # inode's number
    'st_mode' / mode_t,  # inode protection mode
    'st_nlink' / nlink_t,  # number of hard links to the file
    'st_uid' / uid_t,  # user-id of owner
    'st_gid' / gid_t,  # group-id of owner
    'st_rdev' / dev_t,  # device type, for special file inode

    'st_atimespec' / timespec,  # time of last access
    'st_atime' / Computed(this.st_atimespec.tv_sec + (this.st_atimespec.tv_nsec / 10 ** 9)),
    'st_mtimespec' / timespec,  # time of last data modification
    'st_mtime' / Computed(this.st_mtimespec.tv_sec + (this.st_mtimespec.tv_nsec / 10 ** 9)),
    'st_ctimespec' / timespec,  # time of last file status change
    'st_ctime' / Computed(this.st_ctimespec.tv_sec + (this.st_ctimespec.tv_nsec / 10 ** 9)),
    'st_size' / off_t,  # file size, in bytes
    'st_blocks' / blkcnt_t,  # blocks allocated for file
    'st_blksize' / blksize_t,  # optimal blocksize for I/O
    'st_flags' / Int32ul,  # user defined flags for file
    'st_gen' / Int32ul,
    'st_lspare' / Int32sl,
    'st_qspare' / Array(2, Int64sl),
    Padding(24),  # Compiled size is 144
)
stat64 = Struct(
    'st_dev' / dev_t,  # device inode resides on
    'st_mode' / mode_t,  # inode protection mode
    'st_nlink' / nlink_t,  # number of hard links to the file
    'st_ino' / ino64_t,  # inode's number
    'st_uid' / uid_t,  # user-id of owner
    'st_gid' / gid_t,  # group-id of owner
    'st_rdev' / dev_t,  # device type, for special file inode
    Padding(4),

    'st_atimespec' / timespec,  # time of last access
    'st_atime' / Computed(this.st_atimespec.tv_sec + (this.st_atimespec.tv_nsec / 10 ** 9)),
    'st_mtimespec' / timespec,  # time of last data modification
    'st_mtime' / Computed(this.st_mtimespec.tv_sec + (this.st_mtimespec.tv_nsec / 10 ** 9)),
    'st_ctimespec' / timespec,  # time of last file status change
    'st_ctime' / Computed(this.st_ctimespec.tv_sec + (this.st_ctimespec.tv_nsec / 10 ** 9)),
    'st_birthtimespec' / timespec,  # time of file creation(birth)
    'st_btime' / Computed(this.st_birthtimespec.tv_sec + (this.st_birthtimespec.tv_nsec / 10 ** 9)),

    'st_size' / off_t,
    'st_blocks' / blkcnt_t,  # blocks allocated for file
    'st_blksize' / blksize_t,  # optimal blocksize for I/O
    'st_flags' / st_flags,  # blocks allocated for file
    'st_gen' / Int32ul,  # user defined flags for file
    # seems like this value doesn't really exist
    # 'st_lspare' / Int32ul,
    'st_qspare' / Array(2, Int64sl),
)

MFSNAMELEN = 15  # length of fs type name, not inc. nul
MNAMELEN = 90  # length of buffer for returned name

# when _DARWIN_FEATURE_64_BIT_INODE is NOT defined
statfs64 = Struct(
    'f_otype' / Int16sl,  # type of file system (reserved: zero)
    'f_oflags' / Int16sl,  # copy of mount flags (reserved: zero)
    Padding(4),
    'f_bsize' / Int64ul,  # fundamental file system block size
    'f_iosize' / Int64ul,  # optimal transfer block size
    'f_blocks' / Int64ul,  # total data blocks in file system
    'f_bfree' / Int64ul,  # free blocks in fs
    'f_bavail' / Int64ul,  # free blocks avail to non-superuser
    'f_files' / Int64ul,  # total file nodes in file system
    'f_ffree' / Int64ul,  # free file nodes in fs
    'f_fsid' / fsid_t,  # file system id
    'f_owner' / uid_t,  # user that mounted the file system
    Padding(4),
    'f_reserved1' / Int16sl,  # reserved for future use
    'f_type' / Int16sl,  # type of file system (reserved)
    Padding(4),
    'f_flags' / Int64ul,  # copy of mount flags
    'f_reserved2' / Int64ul,  # reserved for future use
    'f_fstypename' / PaddedString(MFSNAMELEN, 'utf8'),  # fs type name
    'f_mntonname' / PaddedString(MNAMELEN, 'utf8'),  # directory on which mounted
    'f_mntfromname' / PaddedString(MNAMELEN, 'utf8'),  # mounted file system
    'f_reserved3' / Int8ul,  # reserved for future use
    'f_reserved4' / Int64ul,  # reserved for future use
)

MFSTYPENAMELEN = 16  # length of fs type name including null
MNAMELEN = MAXPATHLEN

# when _DARWIN_FEATURE_64_BIT_INODE is defined
statfs32 = Struct(
    'f_bsize' / Int32ul,  # fundamental file system block size
    'f_iosize' / Int32sl,  # optimal transfer block size
    'f_blocks' / Int64ul,  # total data blocks in file system
    'f_bfree' / Int64ul,  # free blocks in fs
    'f_bavail' / Int64ul,  # free blocks avail to non-superuser
    'f_files' / Int64ul,  # total file nodes in file system
    'f_ffree' / Int64ul,  # free file nodes in fs
    'f_fsid' / fsid_t,  # file system id
    'f_owner' / uid_t,  # user that mounted the filesystem
    'f_type' / Int32ul,  # type of filesystem
    'f_flags' / Int32ul,  # copy of mount exported flags
    'f_fssubtype' / Int32ul,  # fs sub-type (flavor)
    'f_fstypename' / PaddedString(MFSTYPENAMELEN, 'utf8'),  # fs type name
    'f_mntonname' / PaddedString(MAXPATHLEN, 'utf8'),  # directory on which mounted
    'f_mntfromname' / PaddedString(MAXPATHLEN, 'utf8'),  # mounted filesystem
    'f_reserved' / Int32ul[8],  # For future use
)

PROC_ALL_PIDS = 1

# Flavors for proc_pidinfo()
PROC_PIDLISTFDS = 1
PROC_PIDTASKALLINFO = 2
PROC_PIDTBSDINFO = 3
PROC_PIDTASKINFO = 4
PROC_PIDTHREADINFO = 5
PROC_PIDLISTTHREADS = 6
PROC_PIDREGIONINFO = 7
PROC_PIDREGIONPATHINFO = 8
PROC_PIDVNODEPATHINFO = 9
PROC_PIDTHREADPATHINFO = 10
PROC_PIDPATHINFO = 11
PROC_PIDWORKQUEUEINFO = 12
PROC_PIDT_SHORTBSDINFO = 13
PROC_PIDLISTFILEPORTS = 14
PROC_PIDTHREADID64INFO = 15
PROC_PID_RUSAGE = 16
PROC_PIDUNIQIDENTIFIERINFO = 17
PROC_PIDT_BSDINFOWITHUNIQID = 18
PROC_PIDARCHINFO = 19
PROC_PIDCOALITIONINFO = 20
PROC_PIDNOTEEXIT = 21
PROC_PIDREGIONPATHINFO2 = 22
PROC_PIDREGIONPATHINFO3 = 23

# Flavors for proc_pidfdinfo
PROC_PIDFDVNODEINFO = 1
PROC_PIDFDVNODEPATHINFO = 2
PROC_PIDFDSOCKETINFO = 3
PROC_PIDFDPSEMINFO = 4
PROC_PIDFDPSHMINFO = 5
PROC_PIDFDPIPEINFO = 6
PROC_PIDFDKQUEUEINFO = 7
PROC_PIDFDATALKINFO = 8

# Flavors for proc_pidfileportinfo
PROC_PIDFILEPORTVNODEPATHINFO = 2  # out: vnode_fdinfowithpath
PROC_PIDFILEPORTPSHMINFO = 5  # out: pshm_fdinfo
PROC_PIDFILEPORTPIPEINFO = 6  # out: pipe_fdinfo
# used for proc_setcontrol
PROC_SELFSET_PCONTROL = 1
PROC_SELFSET_THREADNAME = 2
PROC_SELFSET_VMRSRCOWNER = 3
PROC_SELFSET_DELAYIDLESLEEP = 4
# used for proc_dirtycontrol
PROC_DIRTYCONTROL_TRACK = 1
PROC_DIRTYCONTROL_SET = 2
PROC_DIRTYCONTROL_GET = 3
PROC_DIRTYCONTROL_CLEAR = 4
# proc_track_dirty() flags
PROC_DIRTY_TRACK = 0x1
PROC_DIRTY_ALLOW_IDLE_EXIT = 0x2
PROC_DIRTY_DEFER = 0x4
PROC_DIRTY_LAUNCH_IN_PROGRESS = 0x8
# proc_get_dirty() flags
PROC_DIRTY_TRACKED = 0x1
PROC_DIRTY_ALLOWS_IDLE_EXIT = 0x2
PROC_DIRTY_IS_DIRTY = 0x4
PROC_DIRTY_LAUNCH_IS_IN_PROGRESS = 0x8

# Flavors for proc_pidoriginatorinfo
PROC_PIDORIGINATOR_UUID = 0x1
PROC_PIDORIGINATOR_BGSTATE = 0x2
# __proc_info() call numbers
PROC_INFO_CALL_LISTPIDS = 0x1
PROC_INFO_CALL_PIDINFO = 0x2
PROC_INFO_CALL_PIDFDINFO = 0x3
PROC_INFO_CALL_KERNMSGBUF = 0x4
PROC_INFO_CALL_SETCONTROL = 0x5
PROC_INFO_CALL_PIDFILEPORTINFO = 0x6
PROC_INFO_CALL_TERMINATE = 0x7
PROC_INFO_CALL_DIRTYCONTROL = 0x8
PROC_INFO_CALL_PIDRUSAGE = 0x9
PROC_INFO_CALL_PIDORIGINATORINFO = 0xa

# defns of process file desc type
PROX_FDTYPE_ATALK = 0
PROX_FDTYPE_VNODE = 1
PROX_FDTYPE_SOCKET = 2
PROX_FDTYPE_PSHM = 3
PROX_FDTYPE_PSEM = 4
PROX_FDTYPE_KQUEUE = 5
PROX_FDTYPE_PIPE = 6
PROX_FDTYPE_FSEVENTS = 7

pbi_flags_t = FlagsEnum(Int32ul,
                        # pbi_flags values
                        PROC_FLAG_SYSTEM=1,  # System process
                        PROC_FLAG_TRACED=2,  # process currently being traced, possibly by gdb
                        PROC_FLAG_INEXIT=4,  # process is working its way in exit()
                        PROC_FLAG_PPWAIT=8,
                        PROC_FLAG_LP64=0x10,  # 64bit process
                        PROC_FLAG_SLEADER=0x20,  # The process is the session leader
                        PROC_FLAG_CTTY=0x40,  # process has a control tty
                        PROC_FLAG_CONTROLT=0x80,  # Has a controlling terminal
                        PROC_FLAG_THCWD=0x100,  # process has a thread with cwd
                        # process control bits for resource starvation
                        PROC_FLAG_PC_THROTTLE=0x200,
                        # In resource starvation situations, this process is to be throttled
                        PROC_FLAG_PC_SUSP=0x400,  # In resource starvation situations, this process is to be suspended
                        PROC_FLAG_PC_KILL=0x600,  # In resource starvation situations, this process is to be terminated
                        PROC_FLAG_PC_MASK=0x600,
                        # process action bits for resource starvation
                        PROC_FLAG_PA_THROTTLE=0x800,  # The process is currently throttled due to resource starvation
                        PROC_FLAG_PA_SUSP=0x1000,  # The process is currently suspended due to resource starvation
                        PROC_FLAG_PSUGID=0x2000,  # process has set privileges since last exec
                        PROC_FLAG_EXEC=0x4000,  # process has called exec
                        PROC_FLAG_DARWINBG=0x8000,  # process in darwin background
                        PROC_FLAG_EXT_DARWINBG=0x10000,  # process in darwin background - external enforcement
                        PROC_FLAG_IOS_APPLEDAEMON=0x20000,  # Process is apple daemon
                        PROC_FLAG_DELAYIDLESLEEP=0x40000,  # Process is marked to delay idle sleep on disk IO
                        PROC_FLAG_IOS_IMPPROMOTION=0x80000,  # Process is daemon which receives importane donation
                        PROC_FLAG_ADAPTIVE=0x100000,  # Process is adaptive
                        PROC_FLAG_ADAPTIVE_IMPORTANT=0x200000,  # Process is adaptive, and is currently important
                        PROC_FLAG_IMPORTANCE_DONOR=0x400000,  # Process is marked as an importance donor
                        PROC_FLAG_SUPPRESSED=0x800000,  # Process is suppressed
                        PROC_FLAG_APPLICATION=0x1000000,  # Process is an application
                        PROC_FLAG_IOS_APPLICATION=0x1000000,  # Process is an application
                        )

MAXCOMLEN = 16

proc_bsdinfo = Struct(
    'pbi_flags' / pbi_flags_t,  # 64bit emulated etc
    'pbi_status' / Int32ul,
    'pbi_xstatus' / Int32ul,
    'pbi_pid' / Int32ul,
    'pbi_ppid' / Int32ul,
    'pbi_uid' / uid_t,
    'pbi_gid' / gid_t,
    'pbi_ruid' / uid_t,
    'pbi_rgid' / gid_t,
    'pbi_svuid' / uid_t,
    'pbi_svgid' / gid_t,
    'rfu_1' / Int32ul,  # reserved
    '_pbi_comm' / Bytes(MAXCOMLEN),
    'pbi_comm' / Computed(lambda x: x._pbi_comm.split(b'\x00', 1)[0].decode()),
    '_pbi_name' / Bytes(2 * MAXCOMLEN),  # empty if no name is registered
    'pbi_name' / Computed(lambda x: x._pbi_name.split(b'\x00', 1)[0].decode()),
    'pbi_nfiles' / Int32ul,
    'pbi_pgid' / Int32ul,
    'pbi_pjobc' / Int32ul,
    'e_tdev' / Int32ul,  # controlling tty dev
    'e_tpgid' / Int32ul,  # tty process group id
    'pbi_nice' / Int32sl,
    'pbi_start_tvsec' / Int64ul,
    'pbi_start_tvusec' / Int64ul,
)

proc_taskinfo = Struct(
    'pti_virtual_size' / Int64ul,  # virtual memory size (bytes)
    'pti_resident_size' / Int64ul,  # resident memory size (bytes)
    'pti_total_user' / Int64ul,  # total time
    'pti_total_system' / Int64ul,
    'pti_threads_user' / Int64ul,  # existing threads only
    'pti_threads_system' / Int64ul,
    'pti_policy' / Int32sl,  # default policy for new threads
    'pti_faults' / Int32sl,  # number of page faults
    'pti_pageins' / Int32sl,  # number of actual pageins
    'pti_cow_faults' / Int32sl,  # number of copy-on-write faults
    'pti_messages_sent' / Int32sl,  # number of messages sent
    'pti_messages_received' / Int32sl,  # number of messages received
    'pti_syscalls_mach' / Int32sl,  # number of mach system calls
    'pti_syscalls_unix' / Int32sl,  # number of unix system calls
    'pti_csw' / Int32sl,  # number of context switches
    'pti_threadnum' / Int32sl,  # number of threads in the task
    'pti_numrunning' / Int32sl,  # number of running threads
    'pti_priority' / Int32sl,  # task priority
)

proc_taskallinfo = Struct(
    'pbsd' / proc_bsdinfo,
    'ptinfo' / proc_taskinfo,
)

proc_regioninfo = Struct(
    'pri_protection' / Int32ul,
    'pri_max_protection' / Int32ul,
    'pri_inheritance' / Int32ul,
    'pri_flags' / Int32ul,  # shared, external pager, is submap
    'pri_offset' / Int64ul,
    'pri_behavior' / Int32ul,
    'pri_user_wired_count' / Int32ul,
    'pri_user_tag' / Int32ul,
    'pri_pages_resident' / Int32ul,
    'pri_pages_shared_now_private' / Int32ul,
    'pri_pages_swapped_out' / Int32ul,
    'pri_pages_dirtied' / Int32ul,
    'pri_ref_count' / Int32ul,
    'pri_shadow_depth' / Int32ul,
    'pri_share_mode' / Int32ul,
    'pri_private_pages_resident' / Int32ul,
    'pri_shared_pages_resident' / Int32ul,
    'pri_obj_id' / Int32ul,
    'pri_depth' / Int32ul,
    'pri_address' / Int64ul,
    'pri_size' / Int64ul,
)

proc_fdinfo = Struct(
    'proc_fd' / Int32sl,
    'proc_fdtype' / Int32ul,
)

proc_fileinfo = Struct(
    'fi_openflags' / Int32ul,
    'fi_status' / Int32ul,
    'fi_offset' / off_t,
    'fi_guardflags' / Int32ul,
)

# A copy of stat64 with static sized fields.
vinfo_stat = stat64

vnode_info = Struct(
    'vi_stat' / vinfo_stat,
    'vi_type' / Int32sl,
    'vi_pad' / Int32sl,
    'vi_fsid' / fsid_t,
)

vnode_info_path = Struct(
    'vip_vi' / vnode_info,
    '_vip_path' / Bytes(MAXPATHLEN),
    'vip_path' / Computed(lambda x: x._vip_path.split(b'\x00', 1)[0].decode()),
)

vnode_fdinfowithpath = Struct(
    'pfi' / proc_fileinfo,
    'pvip' / vnode_info_path,
)

sockbuf_info = Struct(
    'sbi_cc' / uint32_t,
    'sbi_hiwat' / uint32_t,  # SO_RCVBUF, SO_SNDBUF
    'sbi_mbcnt' / uint32_t,
    'sbi_mbmax' / uint32_t,
    'sbi_lowat' / uint32_t,
    'sbi_flags' / short,
    'sbi_timeo' / short,
)

# TCP Sockets

TSI_T_REXMT = 0  # retransmit
TSI_T_PERSIST = 1  # retransmit persistence
TSI_T_KEEP = 2  # keep alive
TSI_T_2MSL = 3  # 2*msl quiet time timer
TSI_T_NTIMERS = 4


class IpAddressAdapter(Adapter):
    def _decode(self, obj, context, path):
        return ".".join(map(str, obj))

    def _encode(self, obj, context, path):
        return list(map(int, obj.split(".")))


in4in6_addr = Struct(
    'i46a_pad32' / u_int32_t[3],
    'i46a_addr4' / IpAddressAdapter(in_addr),
)

in6_addr = Struct(
    'i46a_pad32' / u_int32_t[3],
    'i46a_addr4' / IpAddressAdapter(in_addr),
)

in_sockinfo = Struct(
    'insi_fport' / Int16ub,
    Padding(2),
    'insi_lport' / Int16ub,
    Padding(2),
    'insi_gencnt' / uint64_t,
    'insi_flags' / uint32_t,
    'insi_flow' / uint32_t,
    'insi_vflag' / uint8_t,
    'insi_ip_ttl' / uint8_t,
    Padding(2),
    'rfu_1' / uint32_t,

    # protocol dependent part
    'insi_faddr' / Union(None,
                         'ina_46' / in4in6_addr,
                         'ina_6' / in6_addr),

    'insi_laddr' / Union(None,
                         'ina_46' / in4in6_addr,
                         'ina_6' / in6_addr),

    'insi_v4' / Struct(
        'in4_tos' / u_char,
    ),

    'insi_v6' / Struct(
        'in6_hlim' / uint8_t,
        'in6_cksum' / Int32sl,
        'in6_ifindex' / u_short,
        'in6_hops' / short,
    ),
)

tcp_sockinfo = Struct(
    'tcpsi_ini' / in_sockinfo,
    'in_sockinfo' / Int32sl,
    'tcpsi_timer' / Int32sl[TSI_T_NTIMERS],
    'tcpsi_mss' / Int32sl,
    'tcpsi_flags' / uint32_t,
    'rfu_1' / uint32_t,  # reserved
    'tcpsi_tp' / uint64_t,  # opaque handle of TCP protocol control block
)

SOCK_MAXADDRLEN = 255

# Unix Domain Sockets

# we can't use sockaddr_un since the sun_path may contain utf8 invalid characters
sockaddr_un_raw = Struct(
    'sun_family' / Default(Int16sl, AF_UNIX),
    '_sun_path' / Bytes(UNIX_PATH_MAX),
    'sun_path' / Computed(lambda x: x._sun_path.split(b'\x00', 1)[0].decode())
)

un_sockinfo = Struct(
    'unsi_conn_so' / uint64_t,
    'unsi_conn_pcb' / uint64_t,
    'unsi_addr' / Union(None,
                        'ua_sun' / sockaddr_un_raw,
                        'ua_dummy' / Bytes(SOCK_MAXADDRLEN)),
    'unsi_caddr' / Union(None,
                         'ua_sun' / sockaddr_un_raw,
                         'ua_dummy' / Bytes(SOCK_MAXADDRLEN)),
)

IF_NAMESIZE = 16

# PF_NDRV Sockets

ndrv_info = Struct(
    'ndrvsi_if_family' / uint32_t,
    'ndrvsi_if_unit' / uint32_t,
    'ndrvsi_if_name' / Bytes(IF_NAMESIZE),
)

# Kernel Event Sockets

kern_event_info = Struct(
    'kesi_vendor_code_filter' / uint32_t,
    'kesi_class_filter' / uint32_t,
    'kesi_subclass_filter' / uint32_t,
)

# Kernel Control Sockets

MAX_KCTL_NAME = 96

kern_ctl_info = Struct(
    'kcsi_id' / uint32_t,
    'kcsi_reg_unit' / uint32_t,
    'kcsi_flags' / uint32_t,
    'kcsi_recvbufsize' / uint32_t,
    'kcsi_sendbufsize' / uint32_t,
    'kcsi_unit' / uint32_t,
    'kcsi_name' / Bytes(MAX_KCTL_NAME),
)

so_kind_t = Enum(Int32ul,
                 SOCKINFO_GENERIC=0,
                 SOCKINFO_IN=1,
                 SOCKINFO_TCP=2,
                 SOCKINFO_UN=3,
                 SOCKINFO_NDRV=4,
                 SOCKINFO_KERN_EVENT=5,
                 SOCKINFO_KERN_CTL=6
                 )

so_family_t = Enum(Int32ul,
                   AF_INET=AF_INET,
                   AF_INET6=AF_INET6,
                   )

socket_info = Struct(
    'soi_stat' / vinfo_stat,
    'soi_so' / uint64_t,  # opaque handle of socket
    'soi_pcb' / uint64_t,  # opaque handle of protocol control block
    'soi_type' / Int32sl,
    'soi_protocol' / Int32sl,
    'soi_family' / so_family_t,
    'soi_options' / short,
    'soi_linger' / short,
    'soi_state' / short,
    'soi_qlen' / short,
    'soi_incqlen' / short,
    'soi_qlimit' / short,
    'soi_timeo' / short,
    'soi_error' / u_short,
    'soi_oobmark' / uint32_t,
    'soi_rcv' / sockbuf_info,
    'soi_snd' / sockbuf_info,
    'soi_kind' / so_kind_t,
    'rfu_1' / uint32_t,  # reserved

    'soi_proto' / Switch(this.soi_kind, {
        so_kind_t.SOCKINFO_IN: Struct('pri_in' / in_sockinfo),
        so_kind_t.SOCKINFO_TCP: Struct('pri_tcp' / tcp_sockinfo),
        so_kind_t.SOCKINFO_UN: Struct('pri_un' / un_sockinfo),
        so_kind_t.SOCKINFO_NDRV: Struct('pri_ndrv' / ndrv_info),
        so_kind_t.SOCKINFO_KERN_EVENT: Struct('pri_kern_event' / kern_event_info),
        so_kind_t.SOCKINFO_KERN_CTL: Struct('pri_kern_ctl' / kern_ctl_info),
    }),
)

socket_fdinfo = Struct(
    'pfi' / proc_fileinfo,
    'psi' / socket_info,
)

pipe_info = Struct(
    'pipe_stat' / vinfo_stat,
    'pipe_handle' / uint64_t,
    'pipe_peerhandle' / uint64_t,
    'pipe_status' / Int32sl,
    'rfu_1' / Int32sl  # reserved
)

pipe_fdinfo = Struct(
    'pfi' / proc_fileinfo,
    'pipeinfo' / pipe_info,
)

vm_region_basic_info = Struct(
    'protection' / vm_prot_t,
    'max_protection' / vm_prot_t,
    'inheritance' / vm_inherit_t,
    'shared' / boolean_t,
    'reserved' / boolean_t,
    Padding(4),
    'offset' / memory_object_offset_t,
    'behavior' / vm_behavior_t,
    'user_wired_count' / Int16ul,
)

vm_region_basic_info_64 = Struct(
    'protection' / vm_prot_t,
    'max_protection' / vm_prot_t,
    'inheritance' / vm_inherit_t,
    'shared' / boolean_t,
    'reserved' / boolean_t,
    Padding(4),
    'offset' / memory_object_offset_t,
    'behavior' / vm_behavior_t,
    'user_wired_count' / Int16ul,
)

VM_REGION_BASIC_INFO_COUNT_64 = vm_region_basic_info_64.sizeof() // 4
vm_region_basic_info_data_t = vm_region_basic_info
natural_t = Int32ul

task_dyld_info = Struct(
    'all_image_info_addr' / mach_vm_address_t,
    'all_image_info_size' / mach_vm_size_t,
    'all_image_info_format' / integer_t,
)

uuid_t = Struct(
    'time_low' / Int64ul,
    'time_mid' / Int32ul,
    'time_hi_and_version' / Int32ul,
    'clock_seq_hi_and_reserved' / Int8ul,
    'clock_seq_low' / Int8ul,
    'node' / Array(6, Int8ul)
)

dyld_image_info_t = Struct(
    'imageLoadAddress' / Int64ul,
    'imageFilePath' / Int64ul,
    'imageFileModDate' / Int64ul
)

all_image_infos_t = Struct(
    'version' / Int32ul,
    'infoArrayCount' / Int32ul,
    'infoArray' / Int64ul,
)

task_dyld_info_data_t = task_dyld_info
TASK_DYLD_INFO_COUNT = task_dyld_info_data_t.sizeof() / natural_t.sizeof()

STRUCT_ARM_THREAD_STATE64 = Struct(
    'x' / Array(29, uint64_t),
    'fp' / Default(uint64_t, 0),
    'lr' / Default(uint64_t, 0),
    'sp' / Default(uint64_t, 0),
    'pc' / Default(uint64_t, 0),
    'cpsr' / Default(uint32_t, 0),
    Padding(4),
)

arm_thread_state64_t = STRUCT_ARM_THREAD_STATE64
ARM_THREAD_STATE64_COUNT = arm_thread_state64_t.sizeof() // uint32_t.sizeof()

STRUCT_X86_THREAD_STATE32 = Struct(
    'eax' / Int32sl,
    'ebx' / Int32sl,
    'ecx' / Int32sl,
    'edx' / Int32sl,
    'edi' / Int32sl,
    'esi' / Int32sl,
    'ebp' / Int32sl,
    'esp' / Int32sl,
    'ss' / Int32sl,
    'eflags' / Int32sl,
    'eip' / Int32sl,
    'cs' / Int32sl,
    'ds' / Int32sl,
    'es' / Int32sl,
    'fs' / Int32sl,
    'gs' / Int32sl,
)

x86_thread_state32_t = STRUCT_X86_THREAD_STATE32

STRUCT_X86_THREAD_STATE64 = Struct(
    'rax' / uint64_t,
    'rbx' / uint64_t,
    'rcx' / uint64_t,
    'rdx' / uint64_t,
    'rdi' / uint64_t,
    'rsi' / uint64_t,
    'rbp' / uint64_t,
    'rsp' / uint64_t,
    'r8' / uint64_t,
    'r9' / uint64_t,
    'r10' / uint64_t,
    'r11' / uint64_t,
    'r12' / uint64_t,
    'r13' / uint64_t,
    'r14' / uint64_t,
    'r15' / uint64_t,
    'rip' / uint64_t,
    'rflags' / uint64_t,
    'cs' / uint64_t,
    'fs' / uint64_t,
    'gs' / uint64_t,
)

x86_thread_state64_t = STRUCT_X86_THREAD_STATE64

suseconds_t = uint32_t

timeval = Struct(
    'tv_sec' / time_t,
    Padding(4),
    'tv_usec' / suseconds_t,
    Padding(4),
)

procargs2_t = Struct(
    'argc' / Int32ul,
    'executable' / Aligned(4, CString('utf8')),
    'argv' / Array(this.argc, CString('utf8')),
    '_environ_apple' / GreedyRange(CString('utf8')),
    'environ_apple' / Computed(lambda ctx: [s for s in ctx._environ_apple if s]),
)

# See: https://opensource.apple.com/source/xnu/xnu-7195.81.3/EXTERNAL_HEADERS/mach-o/loader.h
LC_REQ_DYLD = 0x80000000

LOAD_COMMAND_TYPE = Enum(Int32ul,
                         LC_SEGMENT=0x1,
                         LC_SYMTAB=0x2,
                         LC_SYMSEG=0x3,
                         LC_THREAD=0x4,
                         LC_UNIXTHREAD=0x5,
                         LC_LOADFVMLIB=0x6,
                         LC_IDFVMLIB=0x7,
                         LC_IDENT=0x8,
                         LC_FVMFILE=0x9,
                         LC_PREPAGE=0xa,
                         LC_DYSYMTAB=0xb,
                         LC_LOAD_DYLIB=0xc,
                         LC_ID_DYLIB=0xd,
                         LC_LOAD_DYLINKER=0xe,
                         LC_ID_DYLINKER=0xf,
                         LC_PREBOUND_DYLIB=0x10,
                         LC_ROUTINES=0x11,
                         LC_SUB_FRAMEWORK=0x12,
                         LC_SUB_UMBRELLA=0x13,
                         LC_SUB_CLIENT=0x14,
                         LC_SUB_LIBRARY=0x15,
                         LC_TWOLEVEL_HINTS=0x16,
                         LC_PREBIND_CKSUM=0x17,
                         LC_LOAD_WEAK_DYLIB=(0x18 | LC_REQ_DYLD),
                         LC_SEGMENT_64=0x19,
                         LC_ROUTINES_64=0x1a,
                         LC_UUID=0x1b,
                         LC_RPATH=(0x1c | LC_REQ_DYLD),
                         LC_CODE_SIGNATURE=0x1d,
                         LC_SEGMENT_SPLIT_INFO=0x1e,
                         LC_REEXPORT_DYLIB=0x1f | LC_REQ_DYLD,
                         LC_LAZY_LOAD_DYLIB=0x20,
                         LC_ENCRYPTION_INFO=0x21,
                         LC_DYLD_INFO=0x22,
                         LC_DYLD_INFO_ONLY=(0x22 | LC_REQ_DYLD),
                         LC_LOAD_UPWARD_DYLIB=(0x23 | LC_REQ_DYLD),
                         LC_VERSION_MIN_MACOSX=0x24,
                         LC_VERSION_MIN_IPHONEOS=0x25,
                         LC_FUNCTION_STARTS=0x26,
                         LC_DYLD_ENVIRONMENT=0x27,
                         LC_MAIN=(0x28 | LC_REQ_DYLD),
                         LC_DATA_IN_CODE=0x29,
                         LC_SOURCE_VERSION=0x2A,
                         LC_DYLIB_CODE_SIGN_DRS=0x2B,
                         LC_ENCRYPTION_INFO_64=0x2C,
                         LC_LINKER_OPTION=0x2D,
                         LC_LINKER_OPTIMIZATION_HINT=0x2E,
                         LC_VERSION_MIN_TVOS=0x2F,
                         LC_VERSION_MIN_WATCHOS=0x30,
                         LC_NOTE=0x31,
                         LC_BUILD_VERSION=0x32,
                         LC_DYLD_EXPORTS_TRIE=(0x33 | LC_REQ_DYLD),
                         LC_DYLD_CHAINED_FIXUPS=(0x34 | LC_REQ_DYLD),
                         LC_FILESET_ENTRY=(0x35 | LC_REQ_DYLD)
                         )

# Load commands:
# reference - https://opensource.apple.com/source/xnu/xnu-2050.18.24/EXTERNAL_HEADERS/mach-o/loader.h

FAT_MAGIC = 0xcafebabe
FAT_CIGAM = 0xbebafeca

cpu_type_t = Int32ul
cpu_subtype_t = Int32ul

version_t = BitStruct(
    'major' / BitsInteger(16),
    'minor' / Octet,
    'bug' / Octet
)

segment_command_t = Struct(
    'segname' / PaddedString(16, 'utf8'),
    'vmaddr' / Int64ul,
    'vmsize' / Int64ul,
    'fileoff' / Int64ul,
    'filesize' / Int64ul,
    'maxprot' / Int32ul,
    'initprot' / Int32ul,
    'nsects' / Int32ul,
    'flags' / Int32ul,
)

uuid_command_t = Struct(
    'uuid' / Array(16, Int8ul)
)

build_tool_version = Struct(
    'tool' / Int32ul,
    'version' / Int32ul
)

build_version_command_t = Struct(
    'platform' / Int32ul,
    'minos' / version_t,
    'sdk' / version_t,
    'ntools' / Int32ul,
)

encryption_info_command = Struct(
    'cryptoff' / Int32ul,  # file offset of encrypted range
    'cryptsize' / Int32ul,  # file size of encrypted range
    'cryptid' / Int32ul,  # which encryption system, 0 means not-encrypted yet
)

encryption_info_command_64 = Struct(
    'cryptoff' / Int32ul,  # file offset of encrypted range
    'cryptsize' / Int32ul,  # file size of encrypted range
    'cryptid_offset' / Tell,
    'cryptid' / Int32ul,  # which encryption system, 0 means not-encrypted yet
    '_pad' / Int32ul,
)

load_command_t = Struct(
    '_start' / Tell,
    'cmd' / LOAD_COMMAND_TYPE,
    'cmdsize' / Int32ul,
    '_data_offset' / Tell,
    'data' / Switch(this.cmd, {
        LOAD_COMMAND_TYPE.LC_BUILD_VERSION: build_version_command_t,
        LOAD_COMMAND_TYPE.LC_UUID: uuid_command_t,
        LOAD_COMMAND_TYPE.LC_SEGMENT_64: segment_command_t,
        LOAD_COMMAND_TYPE.LC_ENCRYPTION_INFO_64: encryption_info_command_64,
        LOAD_COMMAND_TYPE.LC_ENCRYPTION_INFO: encryption_info_command,
    }, Bytes(this.cmdsize - (this._data_offset - this._start))),
    Seek(this._start + this.cmdsize),
)

mach_header_t = Struct(
    'load_address' / Tell,
    'magic' / Hex(Int32ul),
    'cputype' / Hex(cpu_type_t),
    'cpusubtype' / Hex(cpu_subtype_t),
    'filetype' / Hex(Int32ul),
    'ncmds' / Hex(Int32ul),
    'sizeofcmds' / Hex(Int32ul),
    'flags' / Hex(Int32ul),
    'reserved' / Hex(Int32ul),
    'load_commands' / LazyArray(this.ncmds, load_command_t),
)

fat_arch = Struct(
    'cputype' / Hex(Int32ub),
    'cpusubtype' / Hex(Int32ub),
    'offset' / Hex(Int32ub),
    'size' / Hex(Int32ub),
    'align' / Hex(Int32ub),
)

fat_header = Struct(
    'magic' / Hex(Int32ul),
    'nfat_arch' / Hex(Int32ub),
    'archs' / Array(this.nfat_arch, fat_arch)
)



from typing import Callable

from cached_property import cached_property

from rpcclient.client import Client
from rpcclient.linux.structs import utsname
from rpcclient.protocol import arch_t


class LinuxClient(Client):
    def __init__(self, sock, sysname: str, arch: arch_t, create_socket_cb: Callable):
        super().__init__(sock, sysname, arch, create_socket_cb)

    @cached_property
    def uname(self):
        with self.safe_calloc(utsname.sizeof()) as uname:
            assert 0 == self.symbols.uname(uname)
            return utsname.parse_stream(uname)



import dataclasses
import time
from enum import IntEnum, IntFlag
from typing import List, Optional

from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import ElementNotFoundError, FirstElementNotFoundError, LastElementNotFoundError, \
    MissingLibraryError, RpcAccessibilityTurnedOffError
from rpcclient.structs.consts import RTLD_NOW


class AXDirection(IntEnum):
    Next = 1
    Previous = 2


class FrameStyle(IntEnum):
    Default = 0
    Blue = 0
    Green = 1
    Yellow = 4
    LightGreen = 5
    Invisible = 6


class AXTraits(IntFlag):
    kAXButtonTrait = 0x01
    kAXLinkTrait = 0x02
    kAXVisitedTrait = 0x100000000
    kAXHeaderTrait = 0x10000
    kAXFooterTrait = 0x4000000
    kAXWebContentTrait = 0x20000
    kAXTextEntryTrait = 0x40000
    kAXImageTrait = 0x4
    kAXSelectedTrait = 0x8
    kAXPlaysSoundTrait = 0x10
    kAXKeyboardKeyTrait = 0x20
    kAXStaticTextTrait = 0x40
    kAXSummaryElementTrait = 0x80
    kAXNotEnabledTrait = 0x100
    kAXPickerElementTrait = 0x80000
    kAXUpdatesFrequentlyTrait = 0x200
    kAXIsEditingTrait = 0x200000
    kAXLaunchIconTrait = 0x400000
    kAXFolderIconTrait = 0x4000000000000
    kAXSearchFieldTrait = 0x400
    kAXStatusBarElementTrait = 0x800000
    kAXAllowsLayoutChangeInStatusBarTrait = 0x400000000000000
    kAXSecureTextFieldTrait = 0x1000000
    kAXBackButtonTrait = 0x8000000
    kAXToggleTrait = 0x20000000000000
    kAXSelectionDismissesItemTrait = 0x80000000
    kAXScrollableTrait = 0x200000000
    kAXSpacerTrait = 0x400000000
    kAXTableIndexTrait = 0x800000000
    kAXMapTrait = 0x1000000000
    kAXTextOperationsAvailableTrait = 0x2000000000
    kAXDraggableTrait = 0x4000000000
    kAXGesturePracticeRegionTrait = 0x8000000000
    kAXPopupButtonTrait = 0x10000000000
    kAXAllowsNativeSlidingTrait = 0x20000000000
    kAXTouchContainerTrait = 0x200000000000
    kAXSupportsZoomTrait = 0x400000000000
    kAXTextAreaTrait = 0x800000000000
    kAXBookContentTrait = 0x1000000000000
    kAXStartsMediaSessionTrait = 0x800
    kAXAdjustableTrait = 0x1000
    kAXMenuItemTrait = 0x10000000000000
    kAXAutoCorrectCandidateTrait = 0x20000000
    kAXDeleteKeyTrait = 0x40000000
    kAXTabButtonTrait = 0x10000000
    kAXIgnoreItemChooserTrait = 0x40000000000000
    kAXAllowsDirectInteractionTrait = 0x2000
    kAXCausesPageTurnTrait = 0x4000
    kAXTabBarTrait = 0x8000
    kAXRadioButtonTrait = 0x100000
    kAXMathEquationTrait = 0x40000000000
    kAXInactiveTrait = 0x2000000
    kAXSupportsTrackingDetailTrait = 0x80000000000000
    kAXAlertTrait = 0x100000000000000
    kAXReadOnlyTrait = 0x8000000000000
    kAXProminentIconTrait = 0x1000000000000000


@dataclasses.dataclass
class CGPoint:
    x: float
    y: float

    def __str__(self) -> str:
        return f'{{{self.x}, {self.y}}}'


@dataclasses.dataclass
class CGSize:
    width: float
    height: float

    def __str__(self) -> str:
        return f'{{{self.width}, {self.height}}}'


@dataclasses.dataclass
class CGRect:
    origin: CGPoint
    size: CGSize

    def __str__(self) -> str:
        return f'{{{self.origin}, {self.size}}}'


class AXElement(DarwinSymbol):
    """
    Wrapper to device's AXElement objective-c object.
    This object was written after reversing XADInspectorManager different methods.
    """

    @property
    def first_element(self):
        """ get first element in hierarchy """
        result = self._element_for_attribute(3000)
        if not result:
            raise FirstElementNotFoundError('failed to get first element in hierarchy')

        if result.ui_element.objc_call('boolWithAXAttribute:', 2046):
            result = result._element_for_attribute(3000)
        return result

    @property
    def last_element(self):
        """ get last element in hierarchy """
        result = self._element_for_attribute(3016)
        if not result:
            raise LastElementNotFoundError('failed to get last element in hierarchy')

        if result.ui_element.objc_call('boolWithAXAttribute:', 2046):
            result = result._element_for_attribute(3016)
        return result

    @property
    def identifier(self) -> str:
        """ get element's identifier """
        return self.objc_call('identifier').py(encoding='utf8')

    @property
    def url(self) -> str:
        """ get element's url """
        return self.objc_call('url').py(encoding='utf8')

    @property
    def path(self) -> DarwinSymbol:
        """ get element's path """
        return self.objc_call('path')

    @property
    def frame(self) -> CGRect:
        """ get element's frame """
        d = self.objc_call('frame', return_raw=True).d
        return CGRect(origin=CGPoint(x=d[0], y=d[1]), size=CGSize(width=d[2], height=d[3]))

    @property
    def label(self) -> Optional[str]:
        """ get element's label (actual displayed text) """
        return self.objc_call('label').py(encoding='utf8')

    @property
    def value(self) -> str:
        """ get element's value (actual set value) """
        return self.objc_call('value').py(encoding='utf8')

    @property
    def bundle_identifier(self) -> str:
        """ get element's bundle identifier """
        return self.objc_call('bundleId').py(encoding='utf8')

    @property
    def pid(self) -> int:
        """ get element's pid """
        return self.objc_call('pid').c_uint16

    @property
    def process_name(self) -> str:
        """ get element's process name """
        return self.objc_call('processName').py()

    @property
    def screen_locked(self) -> bool:
        """ get screen lock state """
        return self.objc_call('isScreenLocked') == 1

    @property
    def is_accessibility_opaque_element_provider(self) -> bool:
        return self.objc_call('isAccessibilityOpaqueElementProvider') != 0

    @property
    def parent(self):
        tmp = self._element_for_attribute(2066)
        if tmp:
            return tmp

        tmp = self._element_for_attribute(2092)
        if tmp:
            return tmp.parent

        return None

    @property
    def ui_element(self) -> DarwinSymbol:
        """ get encapsulated AXUIElement """
        return self.objc_call('uiElement')

    @property
    def traits(self) -> AXTraits:
        """ get current element traits """
        return AXTraits(self.objc_call('traits').c_uint64)

    @property
    def elements(self) -> List['AXElement']:
        """ get all current displayed elements """
        result = []
        elements = self.objc_call('explorerElements')
        for i in range(elements.objc_call('count')):
            result.append(self._client.accessibility.axelement(elements.objc_call('objectAtIndex:', i)))
        return result

    def insert_text(self, text: str) -> None:
        """ insert text into currently editable element """
        self.objc_call('insertText:', self._client.cf(text))

    def delete_text(self) -> None:
        """ delete a character from currently editable element """
        self.objc_call('deleteText')

    def highlight(self) -> None:
        """ draw a frame around the element (replace the old one if existing) """
        frame = self.frame
        self._client.accessibility.draw_frame(frame.origin.x, frame.origin.y, frame.size.width, frame.size.height)

    def scroll_to_visible(self) -> None:
        """ scroll until element becomes fully visible """
        self.objc_call('scrollToVisible')

    def press(self) -> None:
        """ press element """
        self.objc_call('press')

    def long_press(self) -> None:
        """ long press element """
        self.objc_call('longPress')

    def __iter__(self):
        current = self.first_element
        while current:
            yield current
            current = current.next()

    def _element_for_attribute(self, axattribute: int, parameter=None):
        if parameter is None:
            result = self.objc_call('elementForAttribute:', axattribute)
        else:
            result = self.objc_call('elementForAttribute:parameter:', axattribute, parameter)
        return AXElement.create(result, self._client)

    def _next_opaque(self, direction=AXDirection.Next):
        element = self

        if not element.is_accessibility_opaque_element_provider:
            element = self.parent

        if not element:
            return

        element = element._element_for_attribute(95225, self._client.cf([
            direction,
            0,
            self._client.symbols.objc_getClass('NSValue').objc_call('valueWithRange:', 0x7fffffff, 0),
            'AXAudit'
        ]))

        if element:
            ui_element = element.ui_element
            if ui_element and ui_element.objc_call('boolWithAXAttribute:', 2046):
                return element._next_opaque

        return element

    def _next_elements_with_count(self, count: int) -> List['AXElement']:
        elements = self.objc_call('nextElementsWithCount:', count)
        result = []
        for i in range(elements.objc_call('count')):
            result.append(AXElement.create(elements.objc_call('objectAtIndex:', i), self._client))
        return result

    def _previous_elements_with_count(self, count: int) -> List['AXElement']:
        elements = self.objc_call('previousElementsWithCount:', count)
        result = []
        for i in range(elements.objc_call('count')):
            result.append(AXElement.create(elements.objc_call('objectAtIndex:', i), self._client))
        return result

    def _set_assistive_focus(self, focused: bool):
        self.ui_element.objc_call('setAXAttribute:withObject:synchronous:', 2018, self._client.cf({
            'focused': int(focused), 'assistiveTech': 'AXAudit'}), 0)
        parent = self._element_for_attribute(2092)
        if parent:
            parent._set_assistive_focus(focused)

    def next(self, direction=AXDirection.Next, cyclic=False):
        """
        Will get and scroll to the next element in the current view.

        This method was created by reversing [XADInspectorManager _nextElementNavigationInDirection:forElement:]
        so we don't really know much about the used consts.
        """
        next_opaque = self._next_opaque(direction)

        if not self.is_accessibility_opaque_element_provider and next_opaque:
            return next_opaque

        if direction == AXDirection.Next:
            next_or_prev_list = self._next_elements_with_count(1)
        else:
            next_or_prev_list = self._previous_elements_with_count(1)

        if next_or_prev_list:
            result = next_or_prev_list[0]
            if result.is_accessibility_opaque_element_provider:
                focused_element = self._element_for_attribute(95226, self._client.cf('AXAudit'))
                if focused_element:
                    focused_element._set_assistive_focus(False)
                result._set_assistive_focus(False)
                result = result._next_opaque(direction)

            if result and not result.is_accessibility_opaque_element_provider:
                return result

        result = self._next_opaque(direction)
        if result:
            return result

        if not self.is_accessibility_opaque_element_provider:
            parent = self.parent
            if parent:
                return parent.next(direction, cyclic=cyclic)

        if cyclic:
            if direction == AXDirection.Next:
                return self._client.accessibility.primary_app.first_element
            return self._client.accessibility.primary_app.last_element

        return None

    def compare_label(self, label: str, auto_scroll=True, draw_frame=True) -> bool:
        """
        compares given label with the self's

        will auto_scroll into current label if required
        """
        if auto_scroll:
            self.scroll_to_visible()

        if draw_frame:
            self.highlight()

        if self.label == label:
            return True
        return False

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} LABEL:{self.label}>'

    def __str__(self) -> str:
        result = self.label
        return result if result else 'NO LABEL'


class Accessibility:
    """ Accessibility utils """

    def __init__(self, client):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client
        self._load_ax_runtime()
        self._load_accessibility_ui()
        self._ui_client = client.symbols.objc_getClass('AXUIClient').objc_call('alloc').objc_call(
            'initWithIdentifier:serviceBundleName:',
            client.cf('AXAuditAXUIClientIdentifier'),
            client.cf('AXAuditAXUIService'))

    @property
    def primary_app(self) -> AXElement:
        if not self.enabled:
            raise RpcAccessibilityTurnedOffError()
        return self.axelement(self._client.symbols.objc_getClass('AXElement').objc_call('primaryApp'))

    @property
    def enabled(self) -> bool:
        return bool(self._client.symbols._AXSApplicationAccessibilityEnabled() or
                    self._client.symbols._AXSAutomationEnabled())

    @enabled.setter
    def enabled(self, value: bool):
        self._client.symbols._AXSSetAutomationEnabled(int(value))

    def hide_frame(self):
        self.draw_frame(0, 0, 0, 0)

    def set_frame_style(self, value: int):
        self._ui_client.objc_call('sendSynchronousMessage:withIdentifier:error:',
                                  self._client.cf({'frameStyle': value}), 2, 0)

    def draw_frame(self, x: float, y: float, width: float, height: float):
        rect = {'frame': f' {{{{{x},{y}}}, {{{width},{height}}}}}'}
        self._ui_client.objc_call('sendSynchronousMessage:withIdentifier:error:', self._client.cf(rect), 1, 0)

    def wait_for_element_by_label(self, label: str, auto_scroll=True, draw_frame=True, timeout=5,
                                  direction=AXDirection.Next, displayed_only=False) -> AXElement:
        """ busy-wait for an element to become available """
        start = time.time()
        while time.time() - start < timeout:
            try:
                return self._get_element_by_label(label, auto_scroll=auto_scroll, draw_frame=draw_frame,
                                                  direction=direction, displayed_only=displayed_only)
            except ElementNotFoundError:
                pass
            except AttributeError:
                pass
        raise ElementNotFoundError(f'failed to find AXElement by label: "{label}" after waiting for {timeout} seconds '
                                   f'for it to load')

    def _get_element_by_label(self, label: str, auto_scroll=True, draw_frame=True, direction=AXDirection.Next,
                              displayed_only=False) -> AXElement:
        """ get an AXElement by given label """
        if direction == AXDirection.Next:
            element = self.primary_app.first_element
            elements_list = self.primary_app.elements
        elif direction == AXDirection.Previous:
            element = self.primary_app.last_element
            elements_list = reversed(self.primary_app.elements)
        else:
            raise TypeError(f'bad value for: {direction}')

        if displayed_only:
            for element in elements_list:
                if element.compare_label(label, auto_scroll=False, draw_frame=draw_frame):
                    return element
        else:
            while True:
                if element is None:
                    break

                if element.compare_label(label, auto_scroll=auto_scroll, draw_frame=draw_frame):
                    if draw_frame:
                        self.hide_frame()
                    return element

                element = element.next(direction=direction)

        if draw_frame:
            self.hide_frame()

        raise ElementNotFoundError(f'failed to find AXElement by label: "{label}"')

    def press_elements_by_labels(self, labels: List[str], auto_scroll=True, draw_frame=True, timeout=5,
                                 direction: AXDirection = AXDirection.Next, displayed_only=False):
        """
        press a sequence of labels
        :param labels: label list to press
        :param auto_scroll: scroll over the chosen element
        :param draw_frame: draw a frame over the current element
        :param timeout: timeout to wait for each element to appear
        :param direction: the direction of search
        :param displayed_only: search just in the displayed elements
        """
        for label in labels:
            self.wait_for_element_by_label(label, auto_scroll=auto_scroll, draw_frame=draw_frame, timeout=timeout,
                                           direction=direction, displayed_only=displayed_only).press()

            if draw_frame:
                self.hide_frame()

    def _load_ax_runtime(self):
        options = [
            '/System/Library/PrivateFrameworks/AXRuntime.framework/AXRuntime',
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        raise MissingLibraryError('failed to load AXRuntime')

    def _load_accessibility_ui(self):
        options = [
            '/System/Library/PrivateFrameworks/AccessibilityUI.framework/AccessibilityUI',
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        raise MissingLibraryError('failed to load AccessibilityUI')

    def axelement(self, symbol: DarwinSymbol) -> AXElement:
        return AXElement.create(symbol, self._client)



import pytest

pytestmark = pytest.mark.darwin


def test_query_certificates(client):
    assert len(client.keychain.query_certificates()) > 1


def test_query_keys(client):
    assert len(client.keychain.query_keys()) > 1



import logging

from rpcclient.exceptions import BadReturnValueError


class Backlight:
    def __init__(self, client):
        self._client = client

        BrightnessSystemClient = self._client.symbols.objc_getClass('BrightnessSystemClient')
        if not BrightnessSystemClient:
            logging.error('failed to load BrightnessSystemClient class')
        self._brightness = BrightnessSystemClient.objc_call('new')

    @property
    def brightness(self) -> float:
        """ get brightness value in range: 0.0 - 1.0 """
        return self._brightness.objc_call('copyPropertyForKey:', self._client.cf('DisplayBrightness')).py()['Brightness']

    @brightness.setter
    def brightness(self, value: float):
        """ set brightness in range: 0.0 - 1.0 """
        if not self._brightness.objc_call('setProperty:forKey:', self._client.cf(value),
                                          self._client.cf('DisplayBrightness')):
            raise BadReturnValueError('failed to set DisplayBrightness')



from construct import Array, Bytes, Const, Enum, Float64l, Hex, IfThenElse, Int8ul, Int32ul, Int64ul, PaddedString, \
    PascalString, PrefixedArray, Struct, Switch, Union, this

cmd_type_t = Enum(Int32ul,
                  CMD_EXEC=0,
                  CMD_DLOPEN=1,
                  CMD_DLCLOSE=2,
                  CMD_DLSYM=3,
                  CMD_CALL=4,
                  CMD_PEEK=5,
                  CMD_POKE=6,
                  CMD_REPLY_ERROR=7,
                  CMD_REPLY_PEEK=8,
                  CMD_GET_DUMMY_BLOCK=9,
                  CMD_CLOSE=10,
                  CMD_REPLY_POKE=11,
                  CMD_LISTDIR=12,
                  CMD_SHOWOBJECT=13,
                  CMD_SHOWCLASS=14,
                  CMD_GET_CLASS_LIST=15
                  )

arch_t = Enum(Int32ul,
              ARCH_UNKNOWN=0,
              ARCH_ARM64=1,
              )

DEFAULT_PORT = 5910
SERVER_MAGIC_VERSION = 0x88888807
MAGIC = 0x12345678
MAX_PATH_LEN = 1024

protocol_handshake_t = Struct(
    'magic' / Hex(Int32ul),
    'arch' / arch_t,
    'sysname' / PaddedString(256, 'utf8'),
    'machine' / PaddedString(256, 'utf8'),
)

cmd_exec_t = Struct(
    'background' / Int8ul,
    'argv' / PrefixedArray(Int32ul, PascalString(Int32ul, 'utf8')),
    'envp' / PrefixedArray(Int32ul, PascalString(Int32ul, 'utf8')),
)

cmd_dlopen_t = Struct(
    'filename' / PaddedString(MAX_PATH_LEN, 'utf8'),
    'mode' / Int32ul,
)

cmd_dlclose_t = Struct(
    'lib' / Int64ul,
)

cmd_dlsym_t = Struct(
    'lib' / Int64ul,
    'symbol_name' / PaddedString(MAX_PATH_LEN, 'utf8'),
)

argument_type_t = Enum(Int64ul,
                       Integer=0,
                       Double=1)

argument_t = Struct(
    'type' / argument_type_t,
    'value' / IfThenElse(this.type == argument_type_t.Integer, Int64ul, Float64l),
)

cmd_call_t = Struct(
    'address' / Int64ul,
    'va_list_index' / Int64ul,
    'argv' / PrefixedArray(Int64ul, argument_t),
)

cmd_peek_t = Struct(
    'address' / Int64ul,
    'size' / Int64ul,
)

cmd_poke_t = Struct(
    'address' / Int64ul,
    'size' / Int64ul,
    'data' / Bytes(this.size),
)

cmd_dirlist_t = Struct(
    'filename' / PaddedString(MAX_PATH_LEN, 'utf8'),
)

cmd_showobject_t = Struct(
    'address' / Int64ul
)

cmd_showclass_t = Struct(
    'address' / Int64ul
)

listdir_entry_stat_t = Struct(
    'errno' / Int64ul,
    'st_dev' / Int64ul,  # device inode resides on
    'st_mode' / Int64ul,  # inode protection mode
    'st_nlink' / Int64ul,  # number of hard links to the file
    'st_ino' / Int64ul,  # inode's number
    'st_uid' / Int64ul,  # user-id of owner
    'st_gid' / Int64ul,  # group-id of owner
    'st_rdev' / Int64ul,  # device type, for special file inode
    'st_size' / Int64ul,  # file size, in bytes
    'st_blocks' / Int64ul,  # blocks allocated for file
    'st_blksize' / Int64ul,  # optimal blocksize for I/O
    'st_atime' / Int64ul,
    'st_mtime' / Int64ul,
    'st_ctime' / Int64ul,
)

listdir_entry_t = Struct(
    'd_type' / Int64ul,
    'd_namlen' / Int64ul,
    'lstat' / listdir_entry_stat_t,
    'stat' / listdir_entry_stat_t,
)

protocol_message_t = Struct(
    'magic' / Const(MAGIC, Hex(Int32ul)),
    'cmd_type' / cmd_type_t,
    'data' / Switch(this.cmd_type, {
        cmd_type_t.CMD_EXEC: cmd_exec_t,
        cmd_type_t.CMD_DLOPEN: cmd_dlopen_t,
        cmd_type_t.CMD_DLCLOSE: cmd_dlclose_t,
        cmd_type_t.CMD_DLSYM: cmd_dlsym_t,
        cmd_type_t.CMD_CALL: cmd_call_t,
        cmd_type_t.CMD_PEEK: cmd_peek_t,
        cmd_type_t.CMD_POKE: cmd_poke_t,
        cmd_type_t.CMD_LISTDIR: cmd_dirlist_t,
        cmd_type_t.CMD_SHOWOBJECT: cmd_showobject_t,
        cmd_type_t.CMD_SHOWCLASS: cmd_showclass_t,
    })
)

reply_protocol_message_t = Struct(
    'magic' / Const(MAGIC, Int32ul),
    'cmd_type' / cmd_type_t,
)

exec_chunk_type_t = Enum(Int32ul,
                         CMD_EXEC_CHUNK_TYPE_STDOUT=0,
                         CMD_EXEC_CHUNK_TYPE_ERRORCODE=1,
                         )

exec_chunk_t = Struct(
    'chunk_type' / exec_chunk_type_t,
    'size' / Int32ul,
)

return_registers_arm_t = Struct(
    'x' / Array(8, Int64ul),
    'd' / Array(8, Float64l),
)

call_response_t_size = 128

call_response_t = Struct(
    'return_values' / Union(None,
                            'arm_registers' / return_registers_arm_t,
                            'return_value' / Int64ul,
                            ),
)

dummy_block_t = Int64ul



import pytest

from rpcclient.exceptions import ArgumentError


def test_peek(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.peek(peekable, 0x100)


def test_poke(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.poke(peekable, b'a' * 0x100)


def test_peek_invalid_address(client):
    """
    :param rpcclient.client.Client client:
    """
    with pytest.raises(ArgumentError):
        client.peek(0, 0x10)


def test_poke_invalid_address(client):
    """
    :param rpcclient.client.Client client:
    """
    with pytest.raises(ArgumentError):
        client.poke(0, b'a')


def test_get_dummy_block(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.get_dummy_block() != 0


def test_listdir(client):
    """
    :param rpcclient.client.Client client:
    """
    entries = client.listdir('/')
    assert entries[0].d_name == '.'
    assert entries[1].d_name == '..'


def test_calloc(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_calloc(0x100) as zeros:
        assert client.peek(zeros, 0x100) == b'\x00' * 0x100


def test_env_get_set(client):
    """
    :param rpcclient.client.Client client:
    """
    client.setenv('TEST', 'test')
    assert client.getenv('TEST') == 'test'


def test_environ(client):
    """
    :param rpcclient.client.Client client:
    """
    assert len(client.environ) > 0
    for e in client.environ:
        assert '=' in e


def test_reconnect(client):
    """
    :param rpcclient.client.Client client:
    """
    test_listdir(client)
    client.reconnect()
    test_listdir(client)



import ctypes
import logging
from typing import List, Mapping

from rpcclient.allocated import Allocated
from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import BadReturnValueError, RpcClientException
from rpcclient.structs.consts import RTLD_NOW

logger = logging.getLogger(__name__)


class WifiSavedNetwork:
    def __init__(self, client, wifi_manager: DarwinSymbol, network: DarwinSymbol):
        self._client = client
        self._wifi_manager = wifi_manager
        self._network = network

    @property
    def ssid(self) -> bytes:
        return self.get_property('SSID')

    @property
    def bssid(self) -> str:
        return self.get_property('BSSID')

    def get_property(self, name: str):
        return self._client.symbols.WiFiNetworkGetProperty(self._network, self._client.cf(name)).py()

    def forget(self):
        """ forget Wi-Fi network """
        self._client.symbols.WiFiManagerClientRemoveNetwork(self._wifi_manager, self._network)

    def __repr__(self):
        result = f'<{self.__class__.__name__} '
        result += f'SSID:{self.ssid} '
        result += f'BSSID:{self.bssid}'
        result += '>'
        return result


class WifiScannedNetwork:
    def __init__(self, client, interface: DarwinSymbol, network: Mapping):
        self._client = client
        self._interface = interface
        self.network = network

    @property
    def ssid(self) -> bytes:
        return self.network.get('SSID')

    @property
    def bssid(self) -> str:
        return self.network.get('BSSID')

    @property
    def rssi(self) -> int:
        return ctypes.c_int64(self.network['RSSI']).value

    @property
    def channel(self) -> int:
        return self.network['CHANNEL']

    def connect(self, password: str = None):
        """ connect to Wi-Fi network """
        result = self._client.symbols.Apple80211Associate(self._interface, self._client.cf(self.network),
                                                          self._client.cf(password) if password else 0)

        if result:
            raise BadReturnValueError(f'Apple80211Associate() failed with: {result}')

    def disconnect(self):
        if self._client.symbols.Apple80211Disassociate(self._interface):
            raise BadReturnValueError('Apple80211Disassociate() failed')

    def __repr__(self):
        result = f'<{self.__class__.__name__} '
        result += f'SSID:{self.ssid} '
        result += f'BSSID:{self.bssid} '
        result += f'CHANNEL:{self.channel} '
        result += f'RSSI:{self.rssi}'
        result += '>'
        return result


class WifiInterface(Allocated):
    def __init__(self, client, interface, device):
        super().__init__()
        self._client = client
        self._interface = interface
        self._device = device

    def scan(self, options: Mapping = None) -> List[WifiScannedNetwork]:
        """ perform Wi-Fi scan """

        if options is None:
            options = {}

        result = []
        with self._client.safe_malloc(8) as p_found_networks:
            while True:
                scan_result = self._client.symbols.Apple80211Scan(self._interface, p_found_networks,
                                                                  self._client.cf(options))
                if scan_result == 0:
                    break
                elif scan_result == -1:
                    raise BadReturnValueError('Apple80211Scan failed')
                # else, try again

            for network in p_found_networks[0].py():
                result.append(WifiScannedNetwork(self._client, self._interface, network))

        return result

    def disconnect(self):
        """ disconnect from current Wi-Fi network """
        self._client.symbols.Apple80211Disassociate(self._interface)

    def _deallocate(self):
        self._client.symbols.Apple80211Close(self._interface)


class IosWifi:
    """ network utils """

    def __init__(self, client):
        self._client = client
        self._load_wifi_library()

        self._wifi_manager = self._client.symbols.WiFiManagerClientCreate(0, 0)
        if not self._wifi_manager:
            self._client.raise_errno_exception('WiFiManagerClientCreate failed')

    @property
    def saved_networks(self) -> List[WifiSavedNetwork]:
        result = []

        network_list = self._client.symbols.WiFiManagerClientCopyNetworks(self._wifi_manager).py()
        if not network_list:
            return result

        for network in network_list:
            result.append(WifiSavedNetwork(self._client, self._wifi_manager, network))

        return result

    @property
    def interfaces(self) -> List[str]:
        """ get a list of all available wifi interfaces """
        with self._client.safe_malloc(8) as p_interface:
            if self._client.symbols.Apple80211Open(p_interface):
                self._client.raise_errno_exception('Apple80211Open() failed')

            with self._client.safe_malloc(8) as p_interface_names:
                if self._client.symbols.Apple80211GetIfListCopy(p_interface[0], p_interface_names):
                    self._client.raise_errno_exception('Apple80211GetIfListCopy() failed')

                return p_interface_names[0].py()

    def turn_on(self):
        self._set(True)

    def turn_off(self):
        self._set(False)

    def is_on(self) -> bool:
        return bool(self._client.symbols.WiFiManagerClientCopyProperty(self._wifi_manager,
                                                                       self._client.cf('AllowEnable')).py())

    def get_interface(self, interface_name: str = None) -> WifiInterface:
        """ get a specific wifi interface object for remote controlling """
        with self._client.safe_malloc(8) as p_handle:
            if self._client.symbols.Apple80211Open(p_handle):
                self._client.raise_errno_exception('Apple80211Open() failed')
            handle = p_handle[0]

        if interface_name is None:
            wifi_interfaces = self.interfaces

            if not wifi_interfaces:
                raise RpcClientException('no available wifi interfaces were found')

            interface_name = wifi_interfaces[0]

        if self._client.symbols.Apple80211BindToInterface(handle, self._client.cf(interface_name)):
            self._client.raise_errno_exception('Apple80211BindToInterface failed')

        device = self._client.symbols.WiFiManagerClientGetDevice(self._wifi_manager,
                                                                 self._client.cf(interface_name))
        if not device:
            self._client.raise_errno_exception('WiFiManagerClientGetDevice failed')

        return WifiInterface(self._client, handle, device)

    def _load_wifi_library(self):
        options = [
            # macOS
            '/System/Library/Frameworks/CoreWLAN.framework/Versions/A/CoreWLAN',
            '/System/Library/Frameworks/CoreWLAN.framework/CoreWLAN',
            # iOS
            '/System/Library/PrivateFrameworks/WiFiKit.framework/WiFiKit'
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        logger.warning('WiFi library isn\'t available')

    def _set(self, is_on: bool):
        if not self._client.symbols.WiFiManagerClientSetProperty(self._wifi_manager,
                                                                 self._client.cf('AllowEnable'),
                                                                 self._client.cf(is_on)):
            raise BadReturnValueError('WiFiManagerClientSetProperty() failed')



import struct
from enum import IntEnum

from rpcclient.darwin.symbol import DarwinSymbol


class CTL(IntEnum):
    UNSPEC = 0,  # unused
    KERN = 1,  # "high kernel": proc, limits
    VM = 2,  # virtual memory
    VFS = 3,  # file system, mount type is next
    NET = 4,  # network, see socket.h
    DEBUG = 5,  # debugging parameters
    HW = 6,  # generic cpu/io
    MACHDEP = 7,  # machine dependent
    USER = 8,  # user-level
    MAXID = 9,  # number of valid top-level ids


class KERN(IntEnum):
    OSTYPE = 1  # string: system version
    OSRELEASE = 2  # string: system release
    OSREV = 3  # int: system revision
    VERSION = 4  # string: compile time info
    MAXVNODES = 5  # int: max vnodes
    MAXPROC = 6  # int: max processes
    MAXFILES = 7  # int: max open files
    ARGMAX = 8  # int: max arguments to exec
    SECURELVL = 9  # int: system security level
    HOSTNAME = 10  # string: hostname
    HOSTID = 11  # int: host identifier
    CLOCKRATE = 12  # struct: struct clockrate
    VNODE = 13  # struct: vnode structures
    PROC = 14  # struct: process entries
    FILE = 15  # struct: file entries
    PROF = 16  # node: kernel profiling info
    POSIX1 = 17  # int: POSIX.1 version
    NGROUPS = 18  # int: # of supplemental group ids
    JOB_CONTROL = 19  # int: is job control available
    SAVED_IDS = 20  # int: saved set-user/group-ID
    BOOTTIME = 21  # struct: time kernel was booted
    NISDOMAINNAME = 22  # string: YP domain name
    DOMAINNAME = NISDOMAINNAME  # ERN_MAXPARTITIONS = 23  # int: number of partitions/disk
    KDEBUG = 24  # int: kernel trace points
    UPDATEINTERVAL = 25  # int: update process sleep time
    OSRELDATE = 26  # int: OS release date
    NTP_PLL = 27  # node: NTP PLL control
    BOOTFILE = 28  # string: name of booted kernel
    MAXFILESPERPROC = 29  # int: max open files per proc
    MAXPROCPERUID = 30  # int: max processes per uid
    DUMPDEV = 31  # dev_t: device to dump on
    IPC = 32  # node: anything related to IPC
    DUMMY = 33  # unused
    PS_STRINGS = 34  # int: address of PS_STRINGS
    USRSTACK32 = 35  # int: address of USRSTACK
    LOGSIGEXIT = 36  # int: do we log sigexit procs?
    SYMFILE = 37  # string: kernel symbol filename
    PROCARGS = 38  # * 39 was PCSAMPLES... now obsolete
    NETBOOT = 40  # int: are we netbooted? 1=yes,0=no
    # 41 was PANICINFO : panic UI information (deprecated)
    SYSV = 42  # node: System V IPC information
    AFFINITY = 43  # xxx
    TRANSLATE = 44  # xxx
    CLASSIC = TRANSLATE  # XXX backwards compat
    EXEC = 45  # xxx
    CLASSICHANDLER = EXEC  # XXX backwards compatibility
    AIOMAX = 46  # int: max aio requests
    AIOPROCMAX = 47  # int: max aio requests per process
    AIOTHREADS = 48  # int: max aio worker threads
    # ifdef __APPLE_API_UNSTABLE
    PROCARGS2 = 49  # endif # __APPLE_API_UNSTA
    COREFILE = 50  # string: corefile format string
    COREDUMP = 51  # int: whether to coredump at all
    SUGID_COREDUMP = 52  # int: whether to dump SUGID cores
    PROCDELAYTERM = 53  # int: set/reset current proc for delayed termination during shutdown
    SHREG_PRIVATIZABLE = 54  # int: can shared regions be privatized ?
    # 55 was PROC_LOW_PRI_IO... now deprecated
    LOW_PRI_WINDOW = 56  # int: set/reset throttle window - milliseconds
    LOW_PRI_DELAY = 57  # int: set/reset throttle delay - milliseconds
    POSIX = 58  # node: posix tunables
    USRSTACK64 = 59  # LP64 user stack query
    NX_PROTECTION = 60  # int: whether no-execute protection is enabled
    TFP = 61  # Task for pid settings
    PROCNAME = 62  # setup process program  name(2*MAXCOMLEN)
    THALTSTACK = 63  # for compat with older x86 and does nothing
    SPECULATIVE_READS = 64  # int: whether speculative reads are disabled
    OSVERSION = 65  # for build number i.e. 9A127
    SAFEBOOT = 66  # are we booted safe?
    # 67 was LCTX (login context)
    RAGEVNODE = 68  # ERN_TTY = 69  # node: tty settings
    CHECKOPENEVT = 70  # spi: check the VOPENEVT flag on vnodes at open time
    THREADNAME = 71  # set/get thread name
    MAXID = 72  # number of valid kern ids


MAX_SIZE = 0x40000


class Sysctl:
    """ sysctl utils. read man page for sysctl(3) for more details """

    def __init__(self, client):
        self._client = client

    def get(self, ctl: CTL, kern: KERN, arg: int = None, size=MAX_SIZE) -> bytes:
        """ call sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp, void *newp, size_t newlen) on remote """
        namelen = 2

        with self._client.safe_malloc(4 * 3) as mib:
            mib.item_size = 4
            mib[0] = ctl
            mib[1] = kern
            if arg is not None:
                mib[2] = int(arg)
                namelen += 1
            with self._client.safe_malloc(8) as oldenp:
                oldenp[0] = size
                with self._client.safe_malloc(size) as oldp:
                    if self._client.symbols.sysctl(mib, namelen, oldp, oldenp, 0, 0):
                        self._client.raise_errno_exception('sysctl() failed')
                    return oldp.peek(oldenp[0])

    def set(self, ctl: CTL, kern: KERN, oldp: DarwinSymbol, oldenp: DarwinSymbol, arg: int = None):
        """ call sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp, void *newp, size_t newlen) on remote """
        namelen = 2

        with self._client.safe_malloc(4 * 3) as mib:
            mib.item_size = 4
            mib[0] = ctl
            mib[1] = kern
            if arg is not None:
                mib[2] = int(arg)
                namelen += 1
            if self._client.symbols.sysctl(mib, namelen, oldp, oldenp, 0, 0):
                self._client.raise_errno_exception('sysctl() failed')

    def get_str_by_name(self, name: str) -> str:
        """ equivalent of: sysctl <name> """
        return self.get_by_name(name).strip(b'\x00').decode()

    def get_int_by_name(self, name: str) -> int:
        """ equivalent of: sysctl <name> """
        return struct.unpack('<I', self.get_by_name(name))[0]

    def set_int_by_name(self, name: str, value: int):
        """ equivalent of: sysctl <name> -w value """
        self.set_by_name(name, struct.pack('<I', value))

    def set_str_by_name(self, name: str, value: str):
        """ equivalent of: sysctl <name> -w value """
        self.set_by_name(name, value.encode() + b'\x00')

    def set_by_name(self, name: str, value: bytes):
        """ equivalent of: sysctl <name> -w value """
        if self._client.symbols.sysctlbyname(name, 0, 0, value, len(value)):
            self._client.raise_errno_exception('sysctlbyname() failed')

    def get_by_name(self, name: str, size=MAX_SIZE) -> bytes:
        """ equivalent of: sysctl <name> """
        oldval_len = size
        with self._client.safe_malloc(8) as p_oldval_len:
            p_oldval_len[0] = oldval_len
            with self._client.safe_malloc(oldval_len) as oldval:
                if self._client.symbols.sysctlbyname(name, oldval, p_oldval_len, 0, 0):
                    self._client.raise_errno_exception('sysctlbyname() failed')
                return oldval.peek(p_oldval_len[0])



import contextlib
import json
import os
import plistlib
import posixpath
import sys
import tempfile
import time
from argparse import ArgumentParser
from datetime import datetime
from pathlib import Path
from typing import Callable, List, Union
from uuid import UUID

import plumbum
from click.exceptions import Exit
from humanfriendly.prompts import prompt_for_choice
from prompt_toolkit.keys import Keys
from pygments import formatters, highlight, lexers
from pygnuutils.cli.ls import ls as ls_cli
from pygnuutils.ls import Ls, LsStub
from xonsh.built_ins import XSH
from xonsh.cli_utils import Annotated, Arg, ArgParserAlias
from xonsh.events import events
from xonsh.tools import print_color

import rpcclient
from rpcclient.client_factory import create_client
from rpcclient.exceptions import RpcClientException
from rpcclient.structs.consts import SIGTERM


def _default_json_encoder(obj):
    if isinstance(obj, bytes):
        return obj.hex()
    if isinstance(obj, datetime):
        return str(obj)
    if isinstance(obj, UUID):
        return str(obj)
    raise TypeError()


def _pretty_json(buf, colored=True, default=_default_json_encoder):
    formatted_json = json.dumps(buf, sort_keys=True, indent=4, default=default)
    if colored:
        colorful_json = highlight(formatted_json, lexers.JsonLexer(),
                                  formatters.TerminalTrueColorFormatter(style='stata-dark'))
        return colorful_json
    else:
        return formatted_json


def element_completer(xsh, action, completer, alias, command):
    client = XSH.env['rpc']
    result = []
    if not client.accessibility.enabled:
        return result

    for f in client.accessibility.primary_app:
        result.append(f'"{f.label}"')

    return result


def path_completer(xsh, action, completer, alias, command):
    client = XSH.env['rpc']
    pwd = client.fs.pwd()
    is_absolute = command.prefix.startswith('/')
    dirpath = Path(pwd) / command.prefix
    if not client.fs.accessible(dirpath):
        dirpath = dirpath.parent
    result = []
    for f in client.fs.scandir(dirpath):
        if is_absolute:
            completion_option = str((dirpath / f.name))
        else:
            completion_option = str((dirpath / f.name).relative_to(pwd))
        try:
            if f.is_dir():
                result.append(f'{completion_option}/')
            else:
                result.append(completion_option)
        except RpcClientException:
            result.append(completion_option)

    return result


def dir_completer(xsh, action, completer, alias, command):
    client = XSH.env['rpc']
    pwd = client.fs.pwd()
    is_absolute = command.prefix.startswith('/')
    dirpath = Path(pwd) / command.prefix
    if not client.fs.accessible(dirpath):
        dirpath = dirpath.parent
    result = []
    for f in client.fs.scandir(dirpath):
        if is_absolute:
            completion_option = str((dirpath / f.name))
        else:
            completion_option = str((dirpath / f.name).relative_to(pwd))
        try:
            if f.is_dir():
                result.append(f'{completion_option}/')
        except RpcClientException:
            result.append(completion_option)

    return result


class RpcLsStub(LsStub):
    def __init__(self, client, stdout):
        self._client = client
        self._stdout = stdout
        self._paths_entries = {}

    @property
    def sep(self):
        return posixpath.sep

    def join(self, path, *paths):
        return posixpath.join(path, *paths)

    def abspath(self, path):
        return posixpath.normpath(path)

    def stat(self, path, dir_fd=None, follow_symlinks=True):
        path = Path(path)

        for parent, entries in self._paths_entries.items():
            try:
                Path(path).relative_to(parent)
            except ValueError:
                # if not relative then use stat()
                if follow_symlinks:
                    return self._client.fs.stat(path)
                return self._client.fs.lstat(path)

            # otherwise, search for a cached entry
            for e in entries:
                if e.name == path.parts[-1]:
                    return e.stat(follow_symlinks=follow_symlinks)

        return self._client.fs.lstat(path)

    def readlink(self, path, dir_fd=None):
        return self._client.fs.readlink(path)

    def isabs(self, path):
        return posixpath.isabs(path)

    def dirname(self, path):
        return posixpath.dirname(path)

    def basename(self, path):
        return posixpath.basename(path)

    def getgroup(self, st_gid):
        return str(st_gid)

    def getuser(self, st_uid):
        return str(st_uid)

    def now(self):
        return self._client.time.now()

    def listdir(self, path='.'):
        self._paths_entries[path] = self._client.fs.scandir(path)
        return [e.name for e in self._paths_entries[path]]

    def system(self):
        return 'Darwin'

    def getenv(self, key, default=None):
        return self._client.getenv(key)

    def get_tty_width(self):
        return os.get_terminal_size().columns

    def print(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        print(objects[0], end=end, file=self._stdout, flush=flush)


class XonshRc:
    def __init__(self):
        self.client: Union[None, rpcclient.client.Client, rpcclient.darwin.client.DarwinClient] = None
        self._commands = {}
        self._orig_aliases = {}
        self._orig_prompt = XSH.env['PROMPT']
        self._register_rpc_command('rpc-connect', self._rpc_connect)
        self._register_rpc_command('rpc-list-commands', self._rpc_list_commands)
        self._rpc_connect()

        print_color('''
        {BOLD_WHITE}Welcome to xonsh-rpc shell! 👋{RESET}
        Use {CYAN}$rpc{RESET} to access current client.
        Use {CYAN}rpc-list-commands{RESET} to view a list of all available special commands.
            These special commands will replace all already existing commands.

        {BOLD_WHITE}Use the following keyboard shortcuts:{RESET}
        * Home: ControlHome
        * Power: ControlEnd
        * VolUp: ControlShiftUp
        * VolDown: ControlShiftDown
        ''')

    def _register_arg_parse_alias(self, name: str, handler: Callable):
        handler = ArgParserAlias(func=handler, has_args=True, prog=name)
        self._commands[name] = handler
        if XSH.aliases.get(name):
            self._orig_aliases[name] = XSH.aliases[name]
        XSH.aliases[name] = handler

    def _register_rpc_command(self, name, handler):
        self._commands[name] = handler
        if XSH.aliases.get(name):
            self._orig_aliases[name] = XSH.aliases[name]
        XSH.aliases[name] = handler

    def _rpc_disconnect(self, args, stdin, stdout, stderr):
        self.client.close()
        for k, v in self._orig_aliases.items():
            XSH.aliases[k] = v

    def _rpc_connect(self):
        """
        connect to remote rpcserver
        """
        self.client = create_client(XSH.ctx['_create_socket_cb'])

        # -- rpc
        self._register_arg_parse_alias('rpc-disconnect', self._rpc_disconnect)
        self._register_arg_parse_alias('rpc-which', self._rpc_which)

        # -- automation
        self._register_arg_parse_alias('list-elements', self._rpc_list_elements)
        self._register_arg_parse_alias('press-elements', self._rpc_press_elements)
        self._register_arg_parse_alias('enable-accessibility', self._rpc_enable_accessibility)
        self._register_arg_parse_alias('disable-accessibility', self._rpc_disable_accessibility)
        self._register_arg_parse_alias('press-keys', self._rpc_press_keys)
        self._register_arg_parse_alias('swipe-up', self._rpc_swipe_up)
        self._register_arg_parse_alias('swipe-down', self._rpc_swipe_down)
        self._register_arg_parse_alias('swipe-left', self._rpc_swipe_left)
        self._register_arg_parse_alias('swipe-right', self._rpc_swipe_right)

        # -- processes
        self._register_rpc_command('run', self._rpc_run)
        self._register_rpc_command('run-async', self._rpc_run_async)
        if self.client.fs.accessible('/bin/ps'):
            XSH.aliases['ps'] = 'run ps'
        else:
            self._register_arg_parse_alias('ps', self._rpc_ps)
        self._register_arg_parse_alias('kill', self._rpc_kill)
        self._register_arg_parse_alias('killall', self._rpc_killall)

        # -- fs
        self._register_rpc_command('ls', self._rpc_ls)
        self._register_arg_parse_alias('ln', self._rpc_ln)
        self._register_arg_parse_alias('touch', self._rpc_touch)
        self._register_arg_parse_alias('pwd', self._rpc_pwd)
        self._register_arg_parse_alias('cd', self._rpc_cd)
        self._register_arg_parse_alias('rm', self._rpc_rm)
        self._register_arg_parse_alias('mv', self._rpc_mv)
        self._register_arg_parse_alias('cp', self._rpc_cp)
        self._register_arg_parse_alias('mkdir', self._rpc_mkdir)
        self._register_arg_parse_alias('cat', self._rpc_cat)
        self._register_arg_parse_alias('bat', self._rpc_bat)
        self._register_arg_parse_alias('pull', self._rpc_pull)
        self._register_arg_parse_alias('push', self._rpc_push)
        self._register_arg_parse_alias('chmod', self._rpc_chmod)
        self._register_arg_parse_alias('chown', self._rpc_chown)
        self._register_arg_parse_alias('find', self._rpc_find)
        self._register_arg_parse_alias('xattr-get-dict', self._rpc_xattr_get_dict)
        self._register_arg_parse_alias('vim', self._rpc_vim)
        self._register_arg_parse_alias('entitlements', self._rpc_entitlements)

        # -- plist
        self._register_arg_parse_alias('plshow', self._rpc_plshow)

        # -- media
        self._register_arg_parse_alias('record', self._rpc_record)
        self._register_arg_parse_alias('play', self._rpc_play)

        # -- misc
        self._register_arg_parse_alias('open', self._rpc_open)
        self._register_arg_parse_alias('date', self._rpc_date)
        self._register_arg_parse_alias('env', self._rpc_env)
        self._register_arg_parse_alias('rpc-file', self._rpc_file)

        XSH.env['PROMPT'] = f'[{{BOLD_GREEN}}{self.client.uname.nodename}{{RESET}} ' \
                            f'{{BOLD_YELLOW}}{{rpc_cwd}}{{RESET}}]{{prompt_end}} '
        XSH.env['PROMPT_FIELDS']['rpc_cwd'] = self._rpc_cwd
        XSH.env['PROMPT_FIELDS']['prompt_end'] = self._prompt

    def _prompt(self) -> str:
        if len(XSH.history) == 0 or XSH.history[-1].rtn == 0:
            return '{BOLD_GREEN}${RESET}'
        return '{BOLD_RED}${RESET}'

    def _rpc_cwd(self) -> str:
        with self.client.reconnect_lock:
            return self.client.fs.pwd()

    def _rpc_xattr_get_dict(self, filename: Annotated[str, Arg(completer=path_completer)]):
        """
        view file xattributes
        """
        return _pretty_json(self.client.fs.dictxattr(filename))

    def _rpc_vim(self, filename: Annotated[str, Arg(completer=path_completer)]):
        """
        use "vim" to edit the given file
        """
        with self._edit_remotely(filename) as f:
            os.system(f'vim "{f}"')
            self._push(f, filename)

    def _rpc_entitlements(self, filename: Annotated[str, Arg(completer=path_completer)]):
        """
        view file entitlements
        """
        return _pretty_json(self.client.lief.get_entitlements(filename))

    def _rpc_list_elements(self):
        """
        list all labels in current main application')
        """
        for element in self.client.accessibility.primary_app:
            print_color(f'🏷  '
                        f'{{BOLD_WHITE}}Label:{{RESET}} {element.label} '
                        f'{{BOLD_WHITE}}Value:{{RESET}} {element.value} '
                        f'{{RESET}}',
                        file=sys.stdout)

    def _rpc_press_elements(self, label: Annotated[List[str], Arg(nargs='+', completer=element_completer)]):
        """
        press labels list by given order
        """
        self.client.accessibility.press_elements_by_labels(label)

    def _rpc_enable_accessibility(self):
        """
        enable accessibility features
        """
        self.client.accessibility.enabled = True

    def _rpc_disable_accessibility(self):
        """
        disable accessibility features
        """
        self.client.accessibility.enabled = False

    def _rpc_press_keys(self, key: Annotated[
        List[str], Arg(nargs='+',
                       completer=lambda xsh, action, completer, alias, command: ['power', 'home', 'volup', 'voldown',
                                                                                 'mute'])]):
        """
        press key list by given order
        """
        keys_map = {
            'power': self.client.hid.send_power_button_press,
            'home': self.client.hid.send_home_button_press,
            'volup': self.client.hid.send_volume_up_button_press,
            'voldown': self.client.hid.send_volume_down_button_press,
            'mute': self.client.hid.send_mute_button_press,
        }
        for k in key:
            keys_map[k]()

    def _rpc_swipe_up(self):
        """
        swipe up
        """
        self.client.hid.send_swipe_up()

    def _rpc_swipe_down(self):
        """
        swipe down
        """
        self.client.hid.send_swipe_down()

    def _rpc_swipe_left(self):
        """
        swipe left
        """
        self.client.hid.send_swipe_left()

    def _rpc_swipe_right(self):
        """
        swipe right
        """
        self.client.hid.send_swipe_right()

    def _rpc_run(self, args, stdin, stdout, stderr):
        parser = ArgumentParser(prog='run', description='execute a program')
        parser.add_argument('arg', nargs='+')
        args = parser.parse_args(args)
        if not stdin:
            stdin = sys.stdin
        result = self.client.spawn(args.arg, raw_tty=True, stdin=stdin, stdout=stdout)
        return result.error

    def _rpc_run_async(self, args, stdin, stdout, stderr):
        parser = ArgumentParser(prog='run-async', description='execute a program in background')
        parser.add_argument('arg', nargs='+')
        args = parser.parse_args(args)
        print(self.client.spawn(args.arg, raw_tty=False, background=True), file=stdout, flush=True)

    def _rpc_kill(self, pid: int, signal_number: Annotated[int, Arg(nargs='?', default=SIGTERM)]):
        """
        kill a process
        """
        signal_number = int(signal_number)
        pid = int(pid)
        self.client.processes.kill(pid, signal_number)

    def _rpc_killall(self, expression: str, signal_number: Annotated[int, Arg(nargs='?', default=SIGTERM)]):
        """
        killall processes with given basename given as a "contain" expression
        """
        signal_number = int(signal_number)
        for p in self.client.processes.grep(expression):
            p.kill(signal_number)

    def _rpc_ps(self):
        """ list processes """
        for p in self.client.processes.list():
            print_color(f'{p.pid:5} {p.path}', file=sys.stdout, flush=True)

    def _rpc_ls(self, args, stdin, stdout, stderr):
        """ list files """
        try:
            with ls_cli.make_context('ls', args) as ctx:
                files = list(map(self._relative_path, ctx.params.pop('files')))
                files = files if files else [self._rpc_pwd()]
                Ls(RpcLsStub(self.client, stdout))(*files, **ctx.params)
        except Exit:
            pass

    def _rpc_ln(self, src: Annotated[str, Arg(completer=path_completer)],
                dst: Annotated[str, Arg(completer=path_completer)], symlink=False):
        """
        create a link

        Parameters
        ----------
        symlink : -s, --symlink
            create a symlink instead of a hard link
        """
        if symlink:
            self.client.fs.symlink(src, dst)
        else:
            self.client.fs.link(src, dst)

    def _rpc_touch(self, filename: Annotated[str, Arg(completer=path_completer)]):
        """
        create an empty file
        """
        self.client.fs.write_file(filename, b'')

    def _rpc_pwd(self):
        """
        get current working directory
        """
        return self.client.fs.pwd()

    def _rpc_cd(self, path: Annotated[str, Arg(completer=dir_completer)]):
        """
        change directory
        """
        self.client.fs.chdir(path)

    def _rpc_rm(self, path: Annotated[List[str], Arg(nargs='+', completer=path_completer)], recursive=False,
                force=False):
        """
        remove files

        Parameters
        ----------
        recursive : -r, --recursive
            remove recursively
        force : -f, --force
            ignore errors
        """
        for p in path:
            self.client.fs.remove(p, recursive=recursive, force=force)

    def _rpc_mv(self, src: Annotated[str, Arg(completer=path_completer)],
                dst: Annotated[str, Arg(completer=path_completer)]):
        """
        move a file
        """
        self.client.fs.rename(src, dst)

    def _rpc_cp(self, src: Annotated[str, Arg(completer=path_completer)],
                dst: Annotated[str, Arg(completer=path_completer)]):
        """
        copy a file
        """
        self.client.fs.write_file(dst, self.client.fs.read_file(src))

    def _rpc_mkdir(self, filename: Annotated[str, Arg(completer=path_completer)], parents: bool = False,
                   mode: Annotated[str, Arg(nargs='?')] = '777'):
        """
        create a directory

        Parameters
        ----------
        parents : -p, --parents
            Create intermediate directories as required.
        """
        self.client.fs.mkdir(filename, mode=int(mode, 8), parents=parents)

    def _rpc_cat(self, filename: Annotated[List[str], Arg(nargs='+', completer=path_completer)]):
        """
        read a list of files
        """
        buf = b''
        for current in filename:
            with self.client.fs.open(current, 'r') as f:
                buf += f.read()
        buf += b'\n'
        try:
            return buf.decode()
        except UnicodeDecodeError:
            return str(buf)

    def _rpc_bat(self, filename: Annotated[List[str], Arg(completer=path_completer)]):
        """
        "bat" (improved-cat) given file
        """
        with self._remote_file(filename) as f:
            os.system(f'bat "{f}"')

    def _rpc_pull(self, remote: Annotated[str, Arg(completer=path_completer)], local: str):
        """
        pull a file from remote
        """
        return self._pull(remote, local)

    def _rpc_push(self, local: str, remote: Annotated[str, Arg(completer=path_completer)]):
        """
        push a file into remote
        """
        return self._push(local, remote)

    def _rpc_chmod(self, mode: str, filename: Annotated[str, Arg(completer=path_completer)], recursive=False):
        """
        chmod at remote

        Parameters
        ----------
        recursive : -R, --recursive
            remove recursively
        """
        self.client.fs.chmod(filename, int(mode, 8), recursive=recursive)

    def _rpc_chown(self, uid: int, gid: int, filename: Annotated[str, Arg(completer=path_completer)], recursive=False):
        """
        chown at remote

        Parameters
        ----------
        recursive : -r, --recursive
            remove recursively
        """
        uid = int(uid)
        gid = int(gid)
        self.client.fs.chown(filename, uid, gid, recursive=recursive)

    def _rpc_find(self, filename: Annotated[str, Arg(completer=path_completer)], depth=True):
        """ find file recursively """
        for f in self.client.fs.find(filename, topdown=not depth):
            print_color(f, file=sys.stdout, flush=True)

    def _rpc_plshow(self, filename: Annotated[str, Arg(completer=path_completer)]):
        """
        parse and show plist
        """
        with self.client.fs.open(filename, 'r') as f:
            return _pretty_json(plistlib.loads(f.read()))

    def _rpc_record(self, filename: Annotated[str, Arg(completer=path_completer)], duration: int):
        """
        start recording for specified duration
        """
        duration = int(duration)
        with self.client.media.get_recorder(filename) as r:
            r.record()
            time.sleep(duration)
            r.stop()

    def _rpc_play(self, filename: Annotated[str, Arg(completer=path_completer)],
                  duration: Annotated[int, Arg(nargs='?')] = None):
        """
        play file
        """
        with self.client.media.get_player(filename) as r:
            r.play()
            if duration:
                duration = int(duration)
                time.sleep(duration)
            else:
                while r.playing:
                    time.sleep(.1)

    def _rpc_open(self, filename: Annotated[str, Arg(completer=path_completer)]):
        """
        open a file from remote using default program')
        """
        open_ = plumbum.local['open']
        upload_changes = 'Upload changes'
        discard_changes = 'Discard changes'
        with self._edit_remotely(filename) as f:
            open_(f)
            if prompt_for_choice([upload_changes, discard_changes]) == upload_changes:
                self._push(f, filename)

    def _rpc_date(self, new_date: Annotated[str, Arg(nargs='?')] = None):
        """
        get/set date
        """
        if not new_date:
            return self.client.time.now()
        self.client.time.set_current(datetime.fromisoformat(new_date))

    def _rpc_which(self, filename: str):
        """ traverse $PATH to find the first matching executable """
        for p in self.client.getenv('PATH').split(':'):
            abs_filename = (Path(p) / filename).absolute()
            if self.client.fs.accessible((Path(p) / filename).absolute()):
                return abs_filename

    def _rpc_env(self):
        """
        view all environment variables
        """
        return '\n'.join(self.client.environ)

    def _rpc_file(self, filename: str):
        """
        show file type
        """
        file = plumbum.local['file']
        with self._remote_file(filename) as f:
            return file(f)

    def _rpc_list_commands(self):
        """
        list all rpc commands
        """
        buf = ''
        for k, v in self._commands.items():
            buf += f'👾 {k}\n'
        print(buf)

    @contextlib.contextmanager
    def _remote_file(self, remote):
        with tempfile.TemporaryDirectory() as local_dir:
            local = Path(local_dir) / Path(remote).parts[-1]
            if self.client.fs.accessible(remote):
                self._pull(remote, local.absolute())
            try:
                yield local.absolute()
            finally:
                pass

    @contextlib.contextmanager
    def _edit_remotely(self, remote):
        with self._remote_file(remote) as local:
            yield local

    def _relative_path(self, filename):
        return posixpath.join(self._rpc_pwd(), filename)

    def _listdir(self, path: str) -> List[str]:
        return self.client.fs.listdir(path)

    def _pull(self, remote_filename, local_filename):
        self.client.fs.pull(remote_filename, local_filename, onerror=lambda x: None)

    def _push(self, local_filename, remote_filename):
        self.client.fs.push(local_filename, remote_filename, onerror=lambda x: None)


# actual RC contents
XSH.aliases['xontrib']('load z argcomplete coreutils fzf-widgets jedi'.split())
XSH.env['fzf_history_binding'] = "c-r"  # Ctrl+R
XSH.env['fzf_ssh_binding'] = "c-s"  # Ctrl+S
XSH.env['fzf_file_binding'] = "c-t"  # Ctrl+T
XSH.env['fzf_dir_binding'] = "c-g"  # Ctrl+G

rc = XonshRc()
XSH.env['rpc'] = rc.client


@events.on_ptk_create
def custom_keybindings(bindings, **kw):
    @bindings.add(Keys.ControlHome)
    def press_home(event):
        XSH.env['rpc'].hid.send_home_button_press()

    @bindings.add(Keys.ControlEnd)
    def press_power(event):
        XSH.env['rpc'].hid.send_power_button_press()

    @bindings.add(Keys.ControlShiftUp)
    def press_volume_up(event):
        XSH.env['rpc'].hid.send_volume_down_button_press()

    @bindings.add(Keys.ControlShiftDown)
    def press_volume_down(event):
        XSH.env['rpc'].hid.send_volume_up_button_press()



class RpcClientException(Exception):
    """ general exception """
    pass


class InvalidServerVersionMagicError(RpcClientException):
    """ server handshake failed due to an invalid magic """
    pass


class ServerDiedError(RpcClientException):
    """ server became disconnected during an operation """
    pass


class SymbolAbsentError(RpcClientException):
    """ trying to access a symbol which is not exported from any library currently loaded into the server's memory """
    pass


class ProcessSymbolAbsentError(RpcClientException):
    """ trying to access a symbol which is not exported from any library currently loaded into the process's memory """
    pass


class ArgumentError(RpcClientException):
    """ at least one of the supplied arguments for a given function was invalid """
    pass


class BadReturnValueError(RpcClientException):
    """ remote c function returned an error """
    pass


class NoSuchPreferenceError(RpcClientException):
    """ attempt to read a preference data which doesn't exist """
    pass


class CfSerializationError(RpcClientException):
    """ failed to encode/decode a cfobject into/from a python object """
    pass


class SpawnError(RpcClientException):
    """ failed to spawn a child process """
    pass


class FailedToConnectError(RpcClientException):
    """ failed to connect to rpcserver """
    pass


class UnrecognizedSelectorError(RpcClientException):
    """ tried to access a non-existing objc object selector """
    pass


class GettingObjectiveCClassError(RpcClientException):
    """ failed to create an objc class wrapper for a given object """
    pass


class MissingLibraryError(RpcClientException):
    """ a required library could not be found """
    pass


class NoEntitlementsError(RpcClientException):
    """ binary contains no entitlements """
    pass


class HarGlobalNotFoundError(RpcClientException):
    """ Failed to find Harlogger global """
    pass


class ElementNotFoundError(RpcClientException):
    """ Failed to find element """
    pass


class FirstElementNotFoundError(ElementNotFoundError):
    """ Failed to find the first element """
    pass


class LastElementNotFoundError(ElementNotFoundError):
    """ Failed to find the last element """
    pass


class RpcFileExistsError(BadReturnValueError):
    """ RPC version for FileExistsError (errno = EEXIST) """
    pass


class RpcFileNotFoundError(BadReturnValueError):
    """ RPC version for FileNotFoundError (errno = ENOENTRY) """
    pass


class RpcBrokenPipeError(BadReturnValueError):
    """ RPC version for BrokenPipeError (errno = EPIPE) """
    pass


class RpcNotEmptyError(BadReturnValueError):
    """ raised when errno = ENOTEMPTY """
    pass


class RpcIsADirectoryError(BadReturnValueError):
    """ RPC version for IsADirectoryError (errno = ENOTEMPTY) """
    pass


class RpcNotADirectoryError(BadReturnValueError):
    """ RPC version for NotADirectoryError (errno = ENOTDIR) """
    pass


class RpcPermissionError(BadReturnValueError):
    """ RPC version for PermissionError (errno = EPERM) """
    pass


class RpcAccessibilityTurnedOffError(BadReturnValueError):
    """ Application AX and Automation is turned off """
    pass


class RpcFailedToRecordError(BadReturnValueError):
    """ An attempt to record has failed """
    pass


class RpcFailedToPlayError(BadReturnValueError):
    """ An attempt to play has failed """
    pass


class RpcResourceTemporarilyUnavailableError(BadReturnValueError):
    """ raised when errno = EAGAIN """
    pass


class RpcConnectionRefusedError(BadReturnValueError):
    """ RPC version for ConnectionRefusedError (errno = ECONNREFUSED) """
    pass


class NoSuchActivityError(RpcClientException):
    pass


class RpcFailedLaunchingAppError(BadReturnValueError):
    """ Failed to launch application """
    pass


class RpcAppleScriptError(BadReturnValueError):
    """ Failed to execute given AppleScript """
    pass


class RpcXpcError(BadReturnValueError):
    """ XPC-related error """
    pass


class RpcXpcSerializationError(RpcXpcError):
    """ Failed to serialize/deserialize XPC message """
    pass


class RpcSetDeveloperModeError(BadReturnValueError):
    """ Failed to set Developer Mode """
    pass



import sqlite3
from typing import Mapping

from rpcclient.network import Network
from rpcclient.structs.consts import SIGKILL

PINNING_RULED_DB = '/private/var/protected/trustd/pinningrules.sqlite3'


class DarwinNetwork(Network):
    """" Network utils """

    def __init__(self, client):
        super().__init__(client)

    @property
    def proxy_settings(self) -> Mapping:
        return self._client.symbols.CFNetworkCopySystemProxySettings().py()

    def remove_certificate_pinning(self) -> None:
        with self._client.fs.remote_file(PINNING_RULED_DB) as local_db_file:
            # truncate pinning rules
            conn = sqlite3.connect(local_db_file)
            cursor = conn.cursor()
            cursor.execute('DELETE FROM rules')
            conn.commit()
            conn.close()

            # push new db
            self._client.fs.push(local_db_file, PINNING_RULED_DB)

            # restart trustd for changes to take affect
            self._client.processes.get_by_basename('trustd').kill(SIGKILL)

    def flush_dns(self) -> None:
        self._client.processes.get_by_basename('mDNSResponder').kill(SIGKILL)
